import json
from typing import Dict, List, Optional
from defog.util import make_async_post_request
import pandas as pd
import asyncio


async def update_db_schema(self, path_to_csv, dev=False, temp=False):
    """
    Update the DB schema via a CSV
    """
    schema_df = pd.read_csv(path_to_csv).fillna("")
    # check columns
    if not all(
        col in schema_df.columns
        for col in ["table_name", "column_name", "data_type", "column_description"]
    ):
        raise ValueError(
            "The CSV must contain the following columns: table_name, column_name, data_type, column_description"
        )
    schema = {}
    for table_name in schema_df["table_name"].unique():
        schema[table_name] = schema_df[schema_df["table_name"] == table_name][
            ["column_name", "data_type", "column_description"]
        ].to_dict(orient="records")

    payload = {
        "api_key": self.api_key,
        "table_metadata": schema,
        "db_type": self.db_type,
        "dev": dev,
        "temp": temp,
    }

    resp = await make_async_post_request(
        url=f"{self.base_url}/update_metadata", payload=payload
    )
    return resp


async def update_glossary(
    self,
    glossary: str = "",
    customized_glossary: dict = None,
    glossary_compulsory: str = "",
    glossary_prunable_units: List[str] = [],
    dev: bool = False,
):
    """
    Updates the glossary on the defog servers.
    :param glossary: The glossary to be used.
    """
    data = {
        "api_key": self.api_key,
        "glossary": glossary,
        "dev": dev,
        "glossary_compulsory": glossary_compulsory,
        "glossary_prunable_units": glossary_prunable_units,
    }
    if customized_glossary:
        data["customized_glossary"] = customized_glossary
    resp = await make_async_post_request(
        url=f"{self.base_url}/update_glossary", payload=data
    )
    return resp


async def delete_glossary(self, user_type=None, dev=False):
    """
    Deletes the glossary on the defog servers.
    """
    data = {
        "api_key": self.api_key,
        "dev": dev,
    }
    if user_type:
        data["key"] = user_type
    r = await make_async_post_request(
        url=f"{self.base_url}/delete_glossary",
        payload=data,
        return_response_object=True,
    )
    if r.status_code == 200:
        print("Glossary deleted successfully.")
    else:
        error_message = r.json().get("message", "")
        print(f"Glossary deletion failed.\nError message: {error_message}")


async def get_glossary(self, mode="general", dev=False):
    """
    Gets the glossary on the defog servers.
    """
    resp = await make_async_post_request(
        url=f"{self.base_url}/get_glossary",
        payload={"api_key": self.api_key, "dev": dev},
    )
    if mode == "general":
        return resp["glossary"]
    elif mode == "customized":
        return resp["customized_glossary"]


async def get_metadata(self, format="markdown", export_path=None, dev=False):
    """
    Gets the metadata on the defog servers.
    """
    resp = await make_async_post_request(
        url=f"{self.base_url}/get_metadata",
        payload={"api_key": self.api_key, "dev": dev},
    )
    items = []
    for table in resp["table_metadata"]:
        for item in resp["table_metadata"][table]:
            item["table_name"] = table
            items.append(item)
    if format == "markdown":
        return pd.DataFrame(items)[
            ["table_name", "column_name", "data_type", "column_description"]
        ].to_markdown(index=False)
    elif format == "csv":
        if export_path is None:
            export_path = "metadata.csv"
        pd.DataFrame(items)[
            ["table_name", "column_name", "data_type", "column_description"]
        ].to_csv(export_path, index=False)
        print(f"Metadata exported to {export_path}")
        return True
    elif format == "json":
        return resp["table_metadata"]


async def get_feedback(self, n_rows: int = 50, start_from: int = 0):
    """
    Gets the feedback on the defog servers.
    """
    resp = await make_async_post_request(
        url=f"{self.base_url}/get_feedback", payload={"api_key": self.api_key}
    )
    df = pd.DataFrame(resp["data"], columns=resp["columns"])
    df["created_at"] = df["created_at"].apply(lambda x: x[:10])
    for col in ["query_generated", "feedback_text"]:
        df[col] = df[col].fillna("")
        df[col] = df[col].apply(lambda x: x.replace("\n", "\\n"))
    return df.iloc[start_from:].head(n_rows).to_markdown(index=False)


async def get_quota(self) -> Optional[Dict]:
    """
    Get the quota usage for the API key.
    """
    api_key = self.api_key
    r = await make_async_post_request(
        url=f"{self.base_url}/check_api_usage",
        payload={"api_key": api_key},
        return_response_object=True,
    )
    # get status code and return None if not 200
    if r.status_code != 200:
        return None
    return r.json()


async def update_golden_queries(
    self,
    golden_queries: List[Dict] = None,
    golden_queries_path: str = None,
    scrub: bool = True,
    dev: bool = False,
):
    """
    Updates the golden queries on the defog servers.
    :param golden_queries: The golden queries to be used.
    :param golden_queries_path: The path to the golden queries CSV.
    :param scrub: Whether to scrub the golden queries.
    """
    if golden_queries is None and golden_queries_path is None:
        raise ValueError("Please provide either golden_queries or golden_queries_path.")

    if golden_queries is None:
        golden_queries = (
            pd.read_csv(golden_queries_path).fillna("").to_dict(orient="records")
        )

    resp = await make_async_post_request(
        url=f"{self.base_url}/update_golden_queries",
        payload={
            "api_key": self.api_key,
            "golden_queries": golden_queries,
            "scrub": scrub,
            "dev": dev,
        },
    )
    print(
        "Golden queries have been received by the system, and will be processed shortly..."
    )
    print(
        "Once that is done, you should be able to see improved results for your questions."
    )
    return resp


async def delete_golden_queries(
    self,
    golden_queries: dict = None,
    golden_queries_path: str = None,
    all: bool = False,
    dev: bool = False,
):
    """
    Updates the golden queries on the defog servers.
    :param golden_queries: The golden queries to be used.
    :param golden_queries_path: The path to the golden queries CSV.
    :param scrub: Whether to scrub the golden queries.
    """
    if golden_queries is None and golden_queries_path is None and not all:
        raise ValueError(
            "Please provide either golden_queries or golden_queries_path, or set all=True."
        )

    if all:
        resp = await make_async_post_request(
            url=f"{self.base_url}/delete_golden_queries",
            payload={"api_key": self.api_key, "all": True, "dev": dev},
        )
        print("All golden queries have now been deleted.")
    else:
        if golden_queries is None:
            golden_queries = (
                pd.read_csv(golden_queries_path).fillna("").to_dict(orient="records")
            )
        resp = await make_async_post_request(
            url=f"{self.base_url}/update_golden_queries",
            payload={"api_key": self.api_key, "golden_queries": golden_queries},
        )
    return resp


async def get_golden_queries(
    self, format: str = "csv", export_path: str = None, dev: bool = False
):
    """
    Gets the golden queries on the defog servers.
    """
    resp = await make_async_post_request(
        url=f"{self.base_url}/get_golden_queries",
        payload={"api_key": self.api_key, "dev": dev},
    )
    golden_queries = resp["golden_queries"]
    if format == "csv":
        if export_path is None:
            export_path = "golden_queries.csv"
        pd.DataFrame(golden_queries).to_csv(export_path, index=False)
        print(f"{len(golden_queries)} golden queries exported to {export_path}")
        return golden_queries
    elif format == "json":
        if export_path is None:
            export_path = "golden_queries.json"
        # Writing JSON asynchronously
        async with open(export_path, "w") as f:
            await f.write(json.dumps(resp, indent=4))
        print(f"{len(golden_queries)} golden queries exported to {export_path}")
        return golden_queries
    else:
        raise ValueError("format must be either 'csv' or 'json'.")


async def create_table_ddl(
    table_name: str, columns: List[Dict[str, str]], add_exists=True
) -> str:
    """
    Return a DDL statement for creating a table from a list of columns
    `columns` is a list of dictionaries with the following keys:
    - column_name: str
    - data_type: str
    - column_description: str
    """
    md_create = ""
    if add_exists:
        md_create += f"CREATE TABLE IF NOT EXISTS {table_name} (\n"
    else:
        md_create += f"CREATE TABLE {table_name} (\n"
    for i, column in enumerate(columns):
        col_name = column["column_name"]
        # if column name has spaces and hasn't been wrapped in double quotes, wrap it in double quotes
        if " " in col_name and not col_name.startswith('"'):
            col_name = f'"{col_name}"'
        dtype = column["data_type"]
        if i < len(columns) - 1:
            md_create += f"  {col_name} {dtype},\n"
        else:
            # avoid the trailing comma for the last line
            md_create += f"  {col_name} {dtype}\n"
    md_create += ");\n"
    return md_create


async def create_ddl_from_metadata(
    metadata: Dict[str, List[Dict[str, str]]], add_exists=True
) -> str:
    """
    Return a DDL statement for creating tables from metadata
    `metadata` is a dictionary with table names as keys and lists of dictionaries as values.
    Each dictionary in the list has the following keys:
    - column_name: str
    - data_type: str
    - column_description: str
    """
    md_create = ""
    for table_name, columns in metadata.items():
        if "." in table_name:
            # table_name = table_name.split(".", 1)[1]
            schema_name = table_name.split(".")[0]

            md_create += f"CREATE SCHEMA IF NOT EXISTS {schema_name};\n"
        md_create += create_table_ddl(table_name, columns, add_exists=add_exists)
    return md_create


async def create_empty_tables(self, dev: bool = False):
    """
    Create empty tables based on metadata
    """
    metadata = self.get_metadata(format="json", dev=dev)
    if self.db_type == "sqlserver":
        ddl = create_ddl_from_metadata(metadata, add_exists=False)
    else:
        ddl = create_ddl_from_metadata(metadata)

    try:
        if self.db_type == "postgres" or self.db_type == "redshift":
            import asyncpg

            conn = await asyncpg.connect(**self.db_creds)
            await conn.execute(ddl)
            await conn.close()
            return True

        elif self.db_type == "mysql":
            import aiomysql

            conn = await aiomysql.connect(**self.db_creds)
            async with conn.cursor() as cur:
                for statement in ddl.split(";"):
                    await cur.execute(statement)
            await conn.commit()
            await conn.ensure_closed()
            return True

        elif self.db_type == "databricks":
            from databricks import sql

            conn = await asyncio.to_thread(sql.connect, **self.db_creds)
            await asyncio.to_thread(conn.execute, ddl)
            await asyncio.to_thread(conn.commit)
            await asyncio.to_thread(conn.close)
            return True

        elif self.db_type == "snowflake":
            import snowflake.connector

            conn = snowflake.connector.connect(
                user=self.db_creds["user"],
                password=self.db_creds["password"],
                account=self.db_creds["account"],
            )
            cur = conn.cursor()
            cur.execute(
                f"USE WAREHOUSE {self.db_creds['warehouse']}"
            )  # set the warehouse
            for statement in ddl.split(";"):
                cur.execute_async(statement)
                query_id = cur.sfqid
                while conn.is_still_running(conn.get_query_status(query_id)):
                    await asyncio.sleep(1)
            cur.close()
            conn.close()
            return True

        elif self.db_type == "bigquery":
            from google.cloud import bigquery

            client = await asyncio.to_thread(
                bigquery.Client.from_service_account_json,
                self.db_creds["json_key_path"],
            )
            for statement in ddl.split(";"):
                await asyncio.to_thread(client.query, statement)
            return True

        elif self.db_type == "sqlserver":
            import aioodbc

            if self.db_creds["database"] != "":
                connection_string = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={self.db_creds['server']};DATABASE={self.db_creds['database']};UID={self.db_creds['user']};PWD={self.db_creds['password']};TrustServerCertificate=yes;Connection Timeout=120;"
            else:
                connection_string = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={self.db_creds['server']};UID={self.db_creds['user']};PWD={self.db_creds['password']};TrustServerCertificate=yes;Connection Timeout=120;"
            conn = await aioodbc.connect(dsn=connection_string)
            cur = await conn.cursor()
            for statement in ddl.split(";"):
                await cur.execute(statement)
            await conn.commit()
            await cur.close()
            await conn.close()
            return True

        else:
            raise ValueError(f"Unsupported DB type: {self.db_type}")
    except Exception as e:
        print(f"Error: {e}")
        return False



import base64
import json
import os

import functions_framework
from defog import Defog

# DEFOG_CREDS_64 will be set via --set-env-vars by the CLI when deploying to GCP
defog_creds64_str = os.environ.get("DEFOG_CREDS_64")
print(f"defog_creds64_str: {defog_creds64_str}")
defog = Defog(base64creds=defog_creds64_str, save_json=False)
print(defog)


@functions_framework.http
def defog_query_http(request):
    # Set CORS headers for the preflight request
    if request.method == "OPTIONS":
        # Allows GET requests from any origin with the Content-Type
        # header and caches preflight response for an 3600s
        headers = {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Methods": "GET",
            "Access-Control-Allow-Headers": "Content-Type",
            "Access-Control-Max-Age": "3600",
        }
        return ("", 204, headers)

    # Set CORS headers for the main request
    headers = {"Access-Control-Allow-Origin": "*"}
    request_json = request.get_json(force=True)
    question = request_json.get("question")
    hard_filters = request_json.get("hard_filters")
    previous_context = request_json.get("previous_context")

    if not previous_context:
        previous_context = []

    answer = defog.run_query(
        question, hard_filters=hard_filters, previous_context=previous_context
    )
    answer = json.dumps(answer, default=str)
    # this is a dictionary with the keys `columns`, `data`, `previous_context`, `generate_query`, `is_successful`, `reason_for_query`, `suggestion_for_further_questions`
    return (answer, 200, headers)



from defog.util import make_async_post_request


async def check_golden_queries_coverage(self, dev: bool = False):
    """
    Check the number of tables and columns inside the metadata schema that are covered by the golden queries.
    """
    url = f"{self.base_url}/get_golden_queries_coverage"
    payload = {"api_key": self.api_key, "dev": dev}
    return await make_async_post_request(url, payload)


async def check_md_valid(self, dev: bool = False):
    """
    Check if the metadata schema is valid.
    """
    url = f"{self.base_url}/check_md_valid"
    payload = {"api_key": self.api_key, "db_type": self.db_type, "dev": dev}
    return await make_async_post_request(url, payload)


async def check_gold_queries_valid(self, dev: bool = False):
    """
    Check if the golden queries are valid and can be executed on a given database without errors.
    """
    url = f"{self.base_url}/check_gold_queries_valid"
    payload = {"api_key": self.api_key, "db_type": self.db_type, "dev": dev}
    return await make_async_post_request(url, payload)


async def check_glossary_valid(self, dev: bool = False):
    """
    Check if the glossary is valid by verifying if all schema, table, and column names referenced are present in the metadata.
    """
    url = f"{self.base_url}/check_glossary_valid"
    payload = {"api_key": self.api_key, "dev": dev}
    return await make_async_post_request(url, payload)


async def check_glossary_consistency(self, dev: bool = False):
    """
    Check if all logic in the glossary is consistent and coherent.
    """
    url = f"{self.base_url}/check_glossary_consistency"
    payload = {"api_key": self.api_key, "dev": dev}
    return await make_async_post_request(url, payload)



import requests
from defog.util import identify_categorical_columns
from io import StringIO
import pandas as pd
import json
from typing import List, Dict, Union


def generate_postgres_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
    schemas: List[str] = [],
) -> Union[Dict, List, str]:
    """
    Returns the schema of the tables in the database. Keys: column_name, data_type, column_description, custom_type_labels
    If tables is non-empty, we only generate the schema for the mentioned tables in the list.
    If schemas is non-empty, we only generate the schema for the mentioned schemas in the list.
    If return_tables_only is True, we return only the table names as a list.
    If upload is True, we send the schema to the defog servers and generate a CSV.
    If upload is False, we return the schema as a dict.
    If scan is True, we also scan the tables for categorical columns to enhance the column description.
    """
    try:
        import psycopg2
    except ImportError:
        raise ImportError(
            "psycopg2 not installed. Please install it with `pip install psycopg2-binary`."
        )

    conn = psycopg2.connect(**self.db_creds)
    cur = conn.cursor()

    if len(tables) == 0:
        # get all tables
        if len(schemas) > 0:
            for schema in schemas:
                cur.execute(
                    "SELECT table_name FROM information_schema.tables WHERE table_schema = %s;",
                    (schema,),
                )
                if schema == "public":
                    tables += [row[0] for row in cur.fetchall()]
                else:
                    tables += [schema + "." + row[0] for row in cur.fetchall()]
        else:
            excluded_schemas = (
                "information_schema",
                "pg_catalog",
                "pg_toast",
                "pg_temp_1",
                "pg_toast_temp_1",
            )
            cur.execute(
                "SELECT table_name, table_schema FROM information_schema.tables WHERE table_schema NOT IN %s;",
                (excluded_schemas,),
            )
            for row in cur.fetchall():
                if row[1] == "public":
                    tables.append(row[0])
                else:
                    tables.append(f"{row[1]}.{row[0]}")

    if return_tables_only:
        conn.close()
        return tables

    print("Getting schema for each table that you selected...")

    table_columns = {}

    # get the columns for each table
    for table_name in tables:
        if "." in table_name:
            schema, table_name = table_name.split(".", 1)
        else:
            schema = "public"
        cur.execute(
            """
            SELECT 
                CAST(column_name AS TEXT), 
                CAST(
                    CASE 
                        WHEN data_type = 'USER-DEFINED' THEN udt_name
                        ELSE data_type 
                    END AS TEXT
                ) AS type,
                col_description(
                    FORMAT('%%s.%%s', table_schema, table_name)::regclass::oid, 
                    ordinal_position
                ) AS column_description,
                CASE
                    WHEN data_type = 'USER-DEFINED' THEN (
                        SELECT string_agg(enumlabel, ', ')
                        FROM pg_enum
                        WHERE enumtypid = (
                            SELECT oid
                            FROM pg_type
                            WHERE typname = udt_name
                        )
                    )
                    ELSE NULL
                END AS custom_type_labels
            FROM information_schema.columns 
            WHERE table_name::text = %s 
            AND table_schema = %s;
            """,
            (
                table_name,
                schema,
            ),
        )
        rows = cur.fetchall()
        rows = [row for row in rows]
        rows = [
            {
                "column_name": i[0],
                "data_type": i[1],
                "column_description": i[2]
                or ""
                + (
                    " This is an enum column with the possible values: " + i[3]
                    if i[3]
                    else ""
                ),
                "custom_type_labels": i[3].split(", ") if i[3] else [],
            }
            for i in rows
        ]
        if len(rows) > 0:
            if scan:
                rows = identify_categorical_columns(cur, table_name, rows)
            if schema == "public":
                table_columns[table_name] = rows
            else:
                table_columns[schema + "." + table_name] = rows
    conn.close()

    print(
        "Sending the schema to the defog servers and generating column descriptions. This might take up to 2 minutes..."
    )
    if upload:
        # send the schemas dict to the defog servers
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": table_columns,
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return table_columns


def get_postgres_functions(
    self, schemas: List[str] = []
) -> Dict[str, List[Dict[str, str]]]:
    """
    Returns the custom functions and their definitions of the mentioned schemas in the database.
    """
    try:
        import psycopg2
    except ImportError:
        raise ImportError(
            "psycopg2 not installed. Please install it with `pip install psycopg2-binary`."
        )

    conn = psycopg2.connect(**self.db_creds)
    cur = conn.cursor()
    functions = {}

    if len(schemas) == 0:
        schemas = ["public"]

    for schema in schemas:
        cur.execute(
            """
            SELECT
                CAST(p.proname AS TEXT) AS function_name,
                pg_get_functiondef(p.oid) AS function_definition
            FROM pg_proc p
            JOIN pg_namespace n ON n.oid = p.pronamespace
            WHERE n.nspname = %s;
            """,
            (schema,),
        )
        rows = [
            {"function_name": row[0], "function_definition": row[1]}
            for row in cur.fetchall()
            if row[1] is not None
        ]
        if rows:
            functions[schema] = rows
    conn.close()
    return functions


def get_postgres_triggers(
    self, schemas: List[str] = []
) -> Dict[str, List[Dict[str, str]]]:
    try:
        import psycopg2
    except ImportError:
        raise ImportError(
            "psycopg2 not installed. Please install it with `pip install psycopg2-binary`."
        )

    conn = psycopg2.connect(**self.db_creds)
    cur = conn.cursor()
    triggers = {}

    if len(schemas) == 0:
        schemas = ["public"]

    for schema in schemas:
        cur.execute(
            """
            SELECT
                CAST(t.tgname AS TEXT) AS trigger_name,
                pg_get_triggerdef(t.oid) AS trigger_definition
            FROM pg_trigger t
            JOIN pg_class c ON t.tgrelid = c.oid
            JOIN pg_namespace n ON c.relnamespace = n.oid
            WHERE n.nspname = %s
            """,
            (schema,),
        )
        rows = [
            {"trigger_name": row[0], "trigger_definition": row[1]}
            for row in cur.fetchall()
            if row[1] is not None
        ]
        if rows:
            triggers[schema] = rows

    conn.close()
    return triggers


def generate_redshift_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    # when upload is True, we send the schema to the defog servers and generate a CSV
    # when its false, we return the schema as a dict
    try:
        import psycopg2
    except ImportError:
        raise ImportError(
            "psycopg2 not installed. Please install it with `pip install psycopg2-binary`."
        )

    if "schema" not in self.db_creds:
        schema = "public"
        conn = psycopg2.connect(**self.db_creds)
    else:
        schema = self.db_creds["schema"]
        del self.db_creds["schema"]
        conn = psycopg2.connect(**self.db_creds)
    cur = conn.cursor()

    schemas = {}

    if len(tables) == 0:
        # get all tables
        cur.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = %s;",
            (schema,),
        )
        tables = [row[0] for row in cur.fetchall()]

    if return_tables_only:
        conn.close()
        return tables

    print("Getting schema for each table that you selected...")
    # get the schema for each table
    for table_name in tables:
        cur.execute(
            "SELECT CAST(column_name AS TEXT), CAST(data_type AS TEXT) FROM information_schema.columns WHERE table_name::text = %s AND table_schema= %s;",
            (
                table_name,
                schema,
            ),
        )
        rows = cur.fetchall()
        rows = [row for row in rows]
        rows = [{"column_name": i[0], "data_type": i[1]} for i in rows]
        if len(rows) > 0:
            if scan:
                cur.execute(f"SET search_path TO {schema}")
                rows = identify_categorical_columns(cur, table_name, rows)
                cur.close()
            schemas[table_name] = rows

    if upload:
        print(
            "Sending the schema to the defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        # send the schemas dict to the defog servers
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


def generate_mysql_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        import mysql.connector
    except:
        raise Exception("mysql-connector not installed.")

    conn = mysql.connector.connect(**self.db_creds)
    cur = conn.cursor()
    schemas = {}

    if len(tables) == 0:
        # get all tables
        db_name = self.db_creds.get("database", "")
        cur.execute(
            f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{db_name}';"
        )
        tables = [row[0] for row in cur.fetchall()]

    if return_tables_only:
        conn.close()
        return tables

    print("Getting schema for the relevant table in your database...")
    # get the schema for each table
    for table_name in tables:
        cur.execute(
            "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = %s;",
            (table_name,),
        )
        rows = cur.fetchall()
        rows = [row for row in rows]
        rows = [{"column_name": i[0], "data_type": i[1]} for i in rows]
        if scan:
            rows = identify_categorical_columns(cur, table_name, rows)
        if len(rows) > 0:
            schemas[table_name] = rows

    conn.close()

    if upload:
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


def generate_databricks_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        from databricks import sql
    except:
        raise Exception("databricks-sql-connector not installed.")

    conn = sql.connect(**self.db_creds)
    schemas = {}
    with conn.cursor() as cur:
        print("Getting schema for each table that you selected...")
        # get the schema for each table

        if len(tables) == 0:
            # get all tables from databricks
            cur.tables(schema_name=self.db_creds.get("schema", "default"))
            tables = [row.TABLE_NAME for row in cur.fetchall()]

        if return_tables_only:
            conn.close()
            return tables

        for table_name in tables:
            cur.columns(
                schema_name=self.db_creds.get("schema", "default"),
                table_name=table_name,
            )
            rows = cur.fetchall()
            rows = [row for row in rows]
            rows = [
                {"column_name": i.COLUMN_NAME, "data_type": i.TYPE_NAME} for i in rows
            ]
            if scan:
                rows = identify_categorical_columns(cur, table_name, rows)
            if len(rows) > 0:
                schemas[table_name] = rows

    conn.close()

    if upload:
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


def generate_snowflake_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        import snowflake.connector
    except:
        raise Exception("snowflake-connector not installed.")

    conn = snowflake.connector.connect(
        user=self.db_creds["user"],
        password=self.db_creds["password"],
        account=self.db_creds["account"],
    )
    conn.cursor().execute(
        f"USE WAREHOUSE {self.db_creds['warehouse']}"
    )  # set the warehouse

    schemas = {}
    alt_types = {"DATE": "TIMESTAMP", "TEXT": "VARCHAR", "FIXED": "NUMERIC"}
    print("Getting schema for each table that you selected...")
    # get the schema for each table
    if len(tables) == 0:
        # get all tables from Snowflake database
        cur = conn.cursor().execute("SHOW TERSE TABLES;")
        res = cur.fetchall()
        tables = [f"{row[3]}.{row[4]}.{row[1]}" for row in res]

    if return_tables_only:
        conn.close()
        return tables

    for table_name in tables:
        rows = []
        for row in conn.cursor().execute(f"SHOW COLUMNS IN {table_name};"):
            rows.append(row)
        rows = [
            {
                "column_name": i[2],
                "data_type": json.loads(i[3])["type"],
                "column_description": i[8],
            }
            for i in rows
        ]
        for idx, row in enumerate(rows):
            if row["data_type"] in alt_types:
                row["data_type"] = alt_types[row["data_type"]]
            rows[idx] = row
        cur = conn.cursor()
        if scan:
            rows = identify_categorical_columns(cur, table_name, rows)
        cur.close()
        if len(rows) > 0:
            schemas[table_name] = rows

    conn.close()

    if upload:
        print(
            "Sending the schema to the defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


def generate_bigquery_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    scan: bool = True,
    return_tables_only: bool = False,
) -> str:
    try:
        from google.cloud import bigquery
    except:
        raise Exception("google-cloud-bigquery not installed.")

    client = bigquery.Client.from_service_account_json(self.db_creds["json_key_path"])
    project_id = [p.project_id for p in client.list_projects()][0]
    datasets = [dataset.dataset_id for dataset in client.list_datasets()]
    schemas = {}

    if len(tables) == 0:
        # get all tables
        tables = []
        for dataset in datasets:
            tables += [
                f"{project_id}.{dataset}.{table.table_id}"
                for table in client.list_tables(dataset=dataset)
            ]

    if return_tables_only:
        return tables

    print("Getting the schema for each table that you selected...")
    # get the schema for each table
    for table_name in tables:
        table = client.get_table(table_name)
        rows = table.schema
        rows = [{"column_name": i.name, "data_type": i.field_type} for i in rows]
        if len(rows) > 0:
            schemas[table_name] = rows

    client.close()

    if upload:
        print(
            "Sending the schema to Defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


def generate_sqlserver_schema(
    self,
    tables: list,
    upload: bool = True,
    return_format: str = "csv",
    return_tables_only: bool = False,
) -> str:
    try:
        import pyodbc
    except:
        raise Exception("pyodbc not installed.")

    if "database" in self.db_creds and self.db_creds["database"] not in ["", None]:
        connection_string = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={self.db_creds['server']};DATABASE={self.db_creds['database']};UID={self.db_creds['user']};PWD={self.db_creds['password']};TrustServerCertificate=yes;Connection Timeout=120;"
    else:
        connection_string = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={self.db_creds['server']};UID={self.db_creds['user']};PWD={self.db_creds['password']};TrustServerCertificate=yes;Connection Timeout=120;"
    conn = pyodbc.connect(connection_string)

    schemas = {}
    cur = conn.cursor()

    SYSTEM_SCHEMAS = ["sys", "information_schema", "guest", "INFORMATION_SCHEMA"]
    DEFAULT_TABLES = [
        "spt_fallback_db",
        "spt_fallback_dev",
        "spt_fallback_usg",
        "spt_monitor",
        "spt_values",
        "MSreplication_options",
    ]

    if len(tables) == 0:
        # get all tables
        cur.execute(
            "SELECT table_catalog, table_schema, table_name FROM information_schema.tables;"
        )

        tables = []
        for row in cur.fetchall():
            if (
                row.table_schema not in SYSTEM_SCHEMAS
                and row.table_name not in DEFAULT_TABLES
            ):
                if not self.db_creds["database"]:
                    # if database is not specified, we return the table catalog + table schema + table name
                    tables.append(
                        row.table_catalog
                        + "."
                        + row.table_schema
                        + "."
                        + row.table_name
                    )
                else:
                    # if database is specified, we return only the table schema + table name
                    # if the table schema is dbo, we return only the table name
                    if row.table_schema == "dbo":
                        tables.append(row.table_name)
                    else:
                        tables.append(row.table_schema + "." + row.table_name)

    if return_tables_only:
        conn.close()
        return tables

    print("Getting schema for each table in your database...")
    # get the schema for each table
    for orig_table_name in tables:
        # if there are two dots, we have the database name and the schema name
        if orig_table_name.count(".") == 2:
            db_name, schema, table_name = orig_table_name.split(".", 2)
        elif orig_table_name.count(".") == 1:
            schema, table_name = orig_table_name.split(".", 1)
            db_name = self.db_creds["database"]
        else:
            schema = "dbo"
            db_name = self.db_creds["database"]
            table_name = orig_table_name
        cur.execute(
            "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = ? AND table_schema = ? AND table_catalog = ?;",
            table_name,
            schema,
            db_name,
        )
        rows = cur.fetchall()
        rows = [row for row in rows]
        rows = [{"column_name": i[0], "data_type": i[1]} for i in rows]
        if len(rows) > 0:
            schemas[orig_table_name] = rows

    conn.close()
    if upload:
        print(
            "Sending the schema to Defog servers and generating column descriptions. This might take up to 2 minutes..."
        )
        r = requests.post(
            f"{self.base_url}/get_schema_csv",
            json={
                "api_key": self.api_key,
                "schemas": schemas,
                "foreign_keys": [],
                "indexes": [],
            },
            verify=False,
        )
        resp = r.json()
        if "csv" in resp:
            csv = resp["csv"]
            if return_format == "csv":
                pd.read_csv(StringIO(csv)).to_csv("defog_metadata.csv", index=False)
                return "defog_metadata.csv"
            else:
                return csv
        else:
            print(f"We got an error!")
            if "message" in resp:
                print(f"Error message: {resp['message']}")
            print(
                f"Please feel free to open a github issue at https://github.com/defog-ai/defog-python if this a generic library issue, or email support@defog.ai."
            )
    else:
        return schemas


def generate_db_schema(
    self,
    tables: list,
    scan: bool = True,
    upload: bool = True,
    return_tables_only: bool = False,
    return_format: str = "csv",
) -> str:
    if self.db_type == "postgres":
        return self.generate_postgres_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "mysql":
        return self.generate_mysql_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "bigquery":
        return self.generate_bigquery_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "redshift":
        return self.generate_redshift_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "snowflake":
        return self.generate_snowflake_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "databricks":
        return self.generate_databricks_schema(
            tables,
            return_format=return_format,
            scan=scan,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    elif self.db_type == "sqlserver":
        return self.generate_sqlserver_schema(
            tables,
            return_format=return_format,
            upload=upload,
            return_tables_only=return_tables_only,
        )
    else:
        raise ValueError(
            f"Creation of a DB schema for {self.db_type} is not yet supported via the library. If you are a premium user, please contact us at founder@defog.ai so we can manually add it."
        )



import requests


def check_golden_queries_coverage(self, dev: bool = False):
    """
    Check the number of tables and columns inside the metadata schema that are covered by the golden queries.
    """
    try:
        r = requests.post(
            f"{self.base_url}/get_golden_queries_coverage",
            json={"api_key": self.api_key, "dev": dev},
            verify=False,
        )
        resp = r.json()
        return resp
    except Exception as e:
        return {"error": str(e)}


def check_md_valid(self, dev: bool = False):
    """
    Check if the metadata schema is valid.
    """
    try:
        r = requests.post(
            f"{self.base_url}/check_md_valid",
            json={"api_key": self.api_key, "db_type": self.db_type, "dev": dev},
            verify=False,
        )
        resp = r.json()
        return resp
    except Exception as e:
        return {"error": str(e)}


def check_gold_queries_valid(self, dev: bool = False):
    """
    Check if the golden queries are valid, and can actually be executed on a given database without errors
    """
    r = requests.post(
        f"{self.base_url}/check_gold_queries_valid",
        json={"api_key": self.api_key, "db_type": self.db_type, "dev": dev},
        verify=False,
    )
    resp = r.json()
    return resp


def check_glossary_valid(self, dev: bool = False):
    """
    Check if glossary is valid by verifying if all schema, table and column names referenced are present in the metadata.
    """
    r = requests.post(
        f"{self.base_url}/check_glossary_valid",
        json={"api_key": self.api_key, "dev": dev},
        verify=False,
    )
    resp = r.json()
    return resp


def check_glossary_consistency(self, dev: bool = False):
    """
    Check if all logic in the glossary is consistent and coherent.
    """
    r = requests.post(
        f"{self.base_url}/check_glossary_consistency",
        json={"api_key": self.api_key, "dev": dev},
        verify=False,
    )
    resp = r.json()
    return resp



import datetime
import decimal
import json
import os
import re
import shutil
import subprocess
import sys
import time

import pandas as pd
import pwinput
import requests
from prompt_toolkit import prompt

import defog
from defog.util import get_feedback, parse_update

USAGE_STRING = """
Usage: defog <command>

Available commands:
    init\t\t\tSetup defog credentials and your database connection
    gen <table1> <table2>\tSpecify tables to generate schema for
    update <csv>\t\tupdate schema (path to CSV) to defog
    query\t\t\tRun a query
    deploy <gcp|aws>\t\tDeploy a defog server as a cloud function
    quota\t\t\tCheck your API quota limits
    docs\t\t\tPrint documentation
    serve\t\t\tServe a defog server locally
"""

home_dir = os.path.expanduser("~")


def main():
    if len(sys.argv) < 2:
        print(USAGE_STRING)
        sys.exit(1)
    if sys.argv[1] == "init":
        init()
    if sys.argv[1] == "gen":
        gen()
    elif sys.argv[1] == "vet" and sys.argv[2] == "metadata":
        vet_metadata()
    elif sys.argv[1] == "update":
        update()
    elif sys.argv[1] == "glossary":
        glossary()
    elif sys.argv[1] == "query":
        query()
    elif sys.argv[1] == "golden":
        golden()
    elif sys.argv[1] == "deploy":
        deploy()
    elif sys.argv[1] == "quota":
        quota()
    elif sys.argv[1] == "docs":
        # TODO
        raise NotImplementedError("docs not implemented yet")
    elif sys.argv[1] == "serve":
        serve()
    elif sys.argv[1] == "serve-webserver":
        serve_webserver()
    elif sys.argv[1] == "serve-static":
        serve_static()
    else:
        print(f"Unknown command: {sys.argv[1]}")
        print(USAGE_STRING)
        sys.exit(1)


def init():
    """
    Initialize defog by creating a config file at ~/.defog/connection.json
    """
    # print welcome message
    print("Welcome to \033[94mdefog.ai\033[0m!\n")
    # check if .defog/connection.json exists
    # if it does, ask if user wants to overwrite
    # if it doesn't, create it
    filepath = os.path.join(home_dir, ".defog", "connection.json")
    if os.path.exists(filepath):
        print(
            "It looks like you've already initialized defog. Do you want to re-enter your database credentials? (y/n)"
        )
        overwrite = prompt().strip()
        if overwrite.lower() != "y":
            print("We'll keep your existing config. No changes were made.")
            sys.exit(0)
        else:
            print("We'll create a new config file at ~/.defog/connection.json")
    else:
        print("We'll create a new config file at ~/.defog/connection.json")
        if not os.path.exists(os.path.join(home_dir, ".defog")):
            os.mkdir(os.path.join(home_dir, ".defog"))

    # prompt user for defog api key if not in environment variable
    if os.environ.get("DEFOG_API_KEY"):
        print(
            "We found your DEFOG_API_KEY in your environment variables. We'll use that."
        )
        api_key = os.environ.get("DEFOG_API_KEY")
    else:
        print(
            "Please enter your DEFOG_API_KEY. You can get it from https://defog.ai/accounts/dashboard/ and creating an account:"
        )
        api_key = prompt().strip()

    # prompt user for db_type
    print(
        "What database are you using? Available options are: "
        + ", ".join(defog.SUPPORTED_DB_TYPES)
    )
    db_type = prompt().strip()
    db_type = db_type.lower()
    while db_type not in defog.SUPPORTED_DB_TYPES:
        print(
            "Sorry, we don't support that database yet. Available options are: "
            + ", ".join(defog.SUPPORTED_DB_TYPES)
        )
        db_type = prompt().strip()
        db_type = db_type.lower()
    # depending on db_type, prompt user for appropriate db_creds
    if db_type == "postgres" or db_type == "redshift":
        print("Please enter your database host:")
        host = prompt().strip()
        # check if host has http:// or https:// and remove it if it does
        if host.startswith("http://") or host.startswith("https://"):
            host = host.split("://")[1]
        print("Please enter your database port:")
        port = prompt().strip()
        print("Please enter your database name:")
        database = prompt().strip()
        print("Please enter your database user:")
        user = prompt().strip()
        password = pwinput.pwinput(
            prompt="Please enter your database password. This won't be displayed as you're typing for privacy reasons"
        )
        db_creds = {
            "host": host,
            "port": port,
            "database": database,
            "user": user,
            "password": password,
        }

    elif db_type == "mysql":
        print("Please enter your database host:")
        host = prompt().strip()
        print("Please enter your database name:")
        database = prompt().strip()
        print("Please enter your database user:")
        user = prompt().strip()
        print("Please enter your database password:")
        password = pwinput.pwinput(prompt="Please enter your database password:")
        db_creds = {
            "host": host,
            "database": database,
            "user": user,
            "password": password,
        }
    elif db_type == "snowflake":
        print("Please enter your database account:")
        account = prompt().strip()
        print("Please enter your database warehouse:")
        warehouse = prompt().strip()
        print("Please enter your database user:")
        user = prompt().strip()
        password = pwinput.pwinput(prompt="Please enter your database password:")
        db_creds = {
            "account": account,
            "warehouse": warehouse,
            "user": user,
            "password": password,
        }
    elif db_type == "databricks":
        print("Please enter your databricks host:")
        host = prompt().strip()
        token = pwinput.pwinput(prompt="Please enter your databricks token:")
        print("Please add your http_path:")
        http_path = prompt().strip()
        print("Please enter your schema name (this is usually 'default'):")
        schema = prompt().strip()
        db_creds = {
            "server_hostname": host,
            "access_token": token,
            "http_path": http_path,
            "schema": schema,
        }
    elif db_type == "bigquery":
        print("Please enter your bigquery json key's path:")
        json_key_path = prompt().strip()
        db_creds = {
            "json_key_path": json_key_path,
        }
    elif db_type == "sqlserver":
        server = prompt("Please enter your database server host\n").strip()
        database = prompt("Please enter your database name\n").strip()
        user = prompt("Please enter your database user\n").strip()
        password = pwinput.pwinput("Please enter your database password\n")
        db_creds = {
            "server": server,
            "database": database,
            "user": user,
            "password": password,
        }

    # write to filepath and print confirmation
    with open(filepath, "w") as f:
        data = {"api_key": api_key, "db_type": db_type, "db_creds": db_creds}
        json.dump(data, f, indent=4)
    print(f"Your configuration has been saved to {filepath}.")

    # prompt user for tables that they would like to register
    print(
        "We're going to register your tables' schema with defog. Just hit enter to skip if you would like to do this later. You can use `defog gen` to generate a schema for your tables later."
    )
    print(
        "Please enter the names of the tables you would like to register, separated by a space:"
    )
    table_names = prompt().strip()
    table_name_list = re.split(r"\s+", table_names.strip())
    # if input is empty, exit
    if table_name_list == [""]:
        print("No tables were registered. Exiting.")
        sys.exit(0)
    else:
        print(
            "Do you want to automatically scan these tables to determine which column might be categorical? The distinct values in each categorical column will be sent to our server. (y/n)"
        )
        scan_option = prompt().strip()
        if scan_option.lower() == "y" or scan_option.lower() == "yes":
            scan = True
        else:
            scan = False
        df = defog.Defog(api_key=api_key, db_type=db_type, db_creds=db_creds)
        filename = df.generate_db_schema(table_name_list, scan=scan)
        pwd = os.getcwd()
        print(
            "Your schema has been generated and is available as a CSV file in this folder at:\n"
        )
        print(f"\033[1m{pwd}/{filename}\033[0m\n")

    print(
        "You can give us more context about your schema by editing the CSV above. Refer to our cookbook on how to do this here: https://defog.notion.site/Cookbook-for-schema-definitions-1650a6855ea447fdb0be75d39975571b#2ba1d37e243e4da3b8f17590b4a3e4e3.\n\nOnce you're done, you can just hit enter to upload the data in the CSV to Defog. If you would like to exit instead, just enter `exit`."
    )
    upload_option = prompt().strip()
    if upload_option == "exit":
        print("Exiting.")
        sys.exit(0)
    else:
        print(
            "We are now uploading this schema to Defog. This might take up to 30 seconds..."
        )
        resp = df.update_db_schema(filename)
        if resp["status"] == "success":
            print("Your schema has been updated. You're ready to start querying!")
        else:
            print("There was an error updating your schema:")
            print(resp)
            print("Please try again, or contact us at founders@defog.ai")
            sys.exit(1)

    print(
        "You're all set! You can now start querying your database using `defog query`."
    )

    print(
        f"You get started, try entering a sample question here, like how many rows are in the table {table_name_list[0]}?"
    )

    query = prompt("Enter your query here: ")
    while query != "e":
        resp = df.run_query(query, retries=3)
        if not resp["ran_successfully"]:
            print(f"Your query did not run successfully. Please try again.")
        else:
            print("Your question generated the following query:\n")
            print(f"\033[1m{resp['query_generated']}\033[0m\n")
            print("Results:\n")
            # print results in tabular format using 'columns' and 'data' keys
            try:
                print_table(resp["columns"], resp["data"])
            except:
                print(resp)
        query = prompt("Enter another query, or type 'e' to exit: ")

    print("Exiting.")
    sys.exit(0)


def gen():
    """
    Generate a schema for a list of tables and print the link to the schema.
    """
    df = defog.Defog()  # load config from .defog/connection.json
    if len(sys.argv) < 3:
        print(
            "defog gen requires a list of tables to generate. Please enter the names of the tables whose schema you would like to generate, separated by a space:"
        )
        table_names = prompt().strip()
        table_name_list = re.split(r"\s+", table_names.strip())
    else:
        table_name_list = sys.argv[2:]

    if table_name_list == [""] or table_name_list == []:
        print("No tables were registered. Exiting.")
        sys.exit(0)
    print(
        "Do you want to automatically scan these tables to determine which column might be categorical? The distinct values in each column likely to be a categorical column will be sent to our server. (y/n)"
    )
    scan_option = prompt().strip()
    if scan_option.lower() == "y" or scan_option.lower() == "yes":
        scan = True
    else:
        scan = False

    filename = df.generate_db_schema(table_name_list, scan=scan)
    pwd = os.getcwd()
    print(
        "Your schema has been generated and is available at the following CSV file in this folder:\n"
    )
    print(f"\033[1m{pwd}/{filename}\033[0m\n")

    print("We are now uploading this auto-generated schema to Defog.")
    df.update_db_schema(filename)

    print(
        "If you modify the auto-generated schema, please run `defog update <csv_filename>` again to refresh the schema on Defog's servers."
    )


def vet_metadata():
    """
    This function reads the metadata provided by the user, and for each
    row, if it's column description is > 100 characters, it will
    call /shorten_column_description to propose a shorter description.
    Finally, it prompts the user to confirm and save the changes.
    """
    # get csv_path from sys.argv or prompt
    if len(sys.argv) < 4:
        print(
            "defog vet-metadata requires a path to a CSV that contains your Database metadata. Please enter the path to the CSV you would like to vet:"
        )
        csv_path = prompt().strip()
    else:
        csv_path = sys.argv[3]
    while not os.path.exists(csv_path):
        print(f"File {csv_path} not found. Please enter a valid path:")
        csv_path = prompt().strip()

    dfg = defog.Defog()  # load config from .defog/connection.json
    schema_df = pd.read_csv(csv_path).fillna("")
    changes_made = False
    for i, row in schema_df.iterrows():
        column_name = row["column_name"]
        column_description = row["column_description"]
        if len(column_description) > 100:
            print(
                f"Column description for \033[1m{column_name}\033[0m is > 100 characters.\n"
            )
            print(f"Old description: {column_description}\n")
            r = requests.post(
                f"{dfg.base_url}/shorten_column_description",
                json={
                    "api_key": dfg.api_key,
                    "column_name": column_name,
                    "column_description": column_description,
                },
                verify=False,
            )
            resp = r.json()
            new_column_description = resp["column_description"]
            # show both descriptions and prompt user if they want to save
            print(f"Suggested description: {new_column_description}\n")
            reply = prompt(
                "Do you want to save this description? Enter (Y/n) or your proposed edit: "
            )
            if reply.lower() == "y":
                schema_df.at[i, "column_description"] = new_column_description
                changes_made = True
            elif reply.lower() == "n":
                pass
            else:
                schema_df.at[i, "column_description"] = reply
                changes_made = True
    if not changes_made:
        print("Your metadata is fine! No changes were made.")
        sys.exit(0)
    if_save = prompt(
        f"Would you like to save the changes back to {csv_path}? Enter 'y' to save or anything else to exit without saving:"
    )
    if if_save.lower() == "y":
        schema_df.to_csv(csv_path, index=False)
        print(f"Changes saved to {csv_path}")
    else:
        print("Changes not saved.")


def update():
    """
    Update the schema in defog with a new schema using the url provided.
    """
    # check for 3rd arg (url), if not there, prompt user for url
    if len(sys.argv) < 3:
        print(
            "defog update requires a CSV that contains your Database metadata. Please enter the path to the CSV you would like to update:"
        )
        filename = prompt().strip()
    else:
        filename = sys.argv[2]
    # load config from .defog/connection.json
    df = defog.Defog()
    # TODO call vet_schema
    # upload schema to defog
    resp = df.update_db_schema(filename)
    if resp["status"] == "success":
        print("Your schema has been updated. You're ready to start querying!")
    else:
        print("There was an error updating your schema:")
        print(resp)
        print("Please try again, or contact us at founders@defog.ai")


def glossary():
    """
    Checks for the next action in sys.argv and runs the appropriate function.
    Available actions are: update, get, delete
    If the sys.argv is missing an action, prompt the user for the action.
    """
    if len(sys.argv) < 3:
        print(
            "defog glossary requires an action. Please enter 'update', 'get', or 'delete':"
        )
        action = prompt().strip().lower()
    else:
        action = sys.argv[2].lower()
    while action not in ["update", "get", "delete", "exit"]:
        print("Please enter 'update', 'get', 'delete', or 'exit' to exit:")
        action = prompt().strip().lower()

    dfg = defog.Defog()
    if action == "update":
        # get path from sys.argv or prompt
        if len(sys.argv) < 4:
            print(
                "defog glossary update requires a path to a .txt file containing your glossary:"
            )
            path = prompt().strip()
        else:
            path = sys.argv[3]
        while not os.path.exists(path):
            print(f"File {path} not found. Please enter a valid path:")
            path = prompt().strip()
        with open(path, "r") as f:
            glossary = f.read()
        resp = dfg.update_glossary(glossary)
        if resp["status"] == "success":
            print("Your glossary has been updated.")
    elif action == "get":
        glossary = dfg.get_glossary()
        if not glossary:
            print("You don't have a glossary yet.")
            return
        print(f"Your glossary is:\n{glossary}")
        print(
            f"Enter the file path to save the glossary to, or just hit enter for default (glossary.txt):"
        )
        path = prompt().strip()
        if path == "":
            path = "glossary.txt"
        with open(path, "w") as f:
            f.write(glossary)
        print(f"Your glossary has been saved to {path}.")
    elif action == "delete":
        dfg.delete_glossary()


def query():
    """
    Run a query and print the results alongside the generated SQL query.
    """
    df = defog.Defog()  # load config from .defog/connection.json
    if len(sys.argv) < 3:
        print("defog query requires a query. Please enter your query:")
        query = prompt().strip()
    else:
        query = sys.argv[2]

    while len(query) > 300 and query != "e":
        print(
            "Your query is too long. If you have general information about your database,"
            + "you can add it to the glossary to help Defog generate queries more accurately.\n"
            + "You can add a glossary by running `defog glossary update <path/to/glossary>`.\n"
            + "You can also use the `defog golden add` command to add golden queries to help Defog generate queries more accurately."
        )
        query = prompt("Please enter a shorter query, or type 'e' to exit first:")

    user_question = ""
    sql_generated = ""
    while True:
        if query == "e":
            print("Exiting.")
            sys.exit(0)
        elif query == "":
            print("Your query cannot be empty.")
            query = prompt("Enter a query, or type 'e' to exit: ")
        user_question = query
        resp = df.run_query(query, retries=3)
        if not resp["ran_successfully"]:
            if "query_generated" in resp:
                print("Defog generated the following query to answer your question:\n")
                print(f"\033[1m{resp['query_generated']}\033[0m\n")

                print(
                    f"However, your query did not run successfully. The error message generated while running the query on your database was\n\n\033[1m{resp['error_message']}\033[0m\n."
                )

                print(
                    f"If you continue to get these errors, please consider updating the metadata in your schema by editing the google CSV generated and running `defog update <path/to/csv>`, or by updating your glossary.\n"
                )
            else:
                print(
                    f"Defog was unable to generate a query for your question. The error message generated while running the query on your database was\n\n\033[1m{resp.get('error_message')}\033[0m\n."
                )
        else:
            sql_generated = resp.get("query_generated")
            print("Defog generated the following query to answer your question:\n")
            print(f"\033[1m{resp['query_generated']}\033[0m\n")

            reason_for_query = resp.get("reason_for_query")
            if reason_for_query:
                reason_for_query = reason_for_query.replace(". ", "\n\n")

                print("This was its reasoning for generating this query:\n")
                print(f"\033[1m{reason_for_query}\033[0m\n")
            else:
                reason_for_query = ""

            print("Results:\n")
            # print results in tabular format using 'columns' and 'data' keys
            try:
                print_table(resp["columns"], resp["data"])
            except:
                print(resp)

            print()
            get_feedback(
                df.api_key, df.db_type, user_question, sql_generated, df.base_url
            )
        query = prompt("Please enter another query, or type 'e' to exit: ")


def golden():
    """
    Allow the user to get, add, or delete golden queries.
    These are used as references during query generation (akin to k-shot learning),
    and can significantly improve the performance of the query generation model.
    """
    # get action from sys.argv or prompt
    if len(sys.argv) < 3:
        print(
            "defog golden requires an action. Please enter 'get', 'add', or 'delete':"
        )
        action = prompt().strip().lower()
    else:
        action = sys.argv[2].lower()
    while action not in ["get", "add", "delete", "exit"]:
        print("Please enter 'get', 'add', 'delete', or 'exit' to exit:")
        action = prompt().strip().lower()

    dfg = defog.Defog()
    if action == "get":
        # get format from sys.argv or prompt
        if len(sys.argv) < 4:
            print(
                "defog golden get requires an export format. Please enter 'json' or 'csv':"
            )
            format = prompt().strip().lower()
        else:
            format = sys.argv[3].lower()
        while format not in ["json", "csv"]:
            print("Please enter 'json' or 'csv':")
            format = prompt().strip().lower()
        # make request to get golden queries
        print("Getting golden queries...")
        golden_queries = dfg.get_golden_queries(format)
        if golden_queries:
            print(f"\nYou have {len(golden_queries)} golden queries:\n")
        for pair in golden_queries:
            question = pair["question"]
            sql = pair["sql"]
            print(f"Question:\n{question}\nSQL:\n{sql}\n")
    elif action == "add":
        # get path from sys.argv or prompt
        if len(sys.argv) < 4:
            print(
                "defog golden add requires a path to a JSON or CSV file containing golden queries:"
            )
            path = prompt().strip()
        else:
            path = sys.argv[3]
        while (
            not os.path.exists(path)
            and not path.endswith(".json")
            and not path.endswith(".csv")
        ):
            print("File not found. Please enter a valid path:")
            path = prompt().strip()
        # if path ends with json, read in json and pass in golden_queries
        if path.endswith(".json"):
            with open(path, "r") as f:
                golden_queries = json.load(f)
                dfg.update_golden_queries(golden_queries=golden_queries)
        # if path ends with csv, pass in csv file
        elif path.endswith(".csv"):
            dfg.update_golden_queries(golden_queries_path=path)
    elif action == "delete":
        # get path from sys.argv or prompt
        if len(sys.argv) < 4:
            print(
                "defog golden delete requires either a path to a JSON or CSV file containing golden queries for deletion, or 'all' to delete all golden queries."
            )
            path = prompt().strip()
        else:
            path = sys.argv[3]
        while (
            not os.path.exists(path)
            and not path.endswith(".json")
            and not path.endswith(".csv")
            and path != "all"
        ):
            print(
                "File not found. Please enter a valid path, or 'all' to delete all golden queries:"
            )
            path = prompt().strip()
        if path == "all":
            dfg.delete_golden_queries(all=True)
        elif path.endswith(".json"):
            with open(path, "r") as f:
                golden_queries = json.load(f)
                dfg.delete_golden_queries(golden_queries=golden_queries)
        elif path.endswith(".csv"):
            dfg.delete_golden_queries(golden_queries_path=path)


def deploy():
    """
    Deploy a cloud function that can be used to run queries.
    """
    # check args for gcp or aws
    if len(sys.argv) < 3:
        print("defog deploy requires a cloud provider. Please enter 'gcp' or 'aws':")
        cloud_provider = prompt().strip().lower()
    else:
        cloud_provider = sys.argv[2].lower()

    if len(sys.argv) >= 4:
        function_name = sys.argv[3]
    else:
        function_name = f"defog-{cloud_provider}"

    # load config from .defog/connection.json
    df = defog.Defog()

    if cloud_provider == "gcp":
        # base64 encode defog credentials for ease of passing around in cli
        creds64_str = df.to_base64_creds()
        source_path = os.path.join(defog.__path__[0], "gcp")
        cmd = [
            "gcloud",
            "functions",
            "deploy",
            function_name,
            "--runtime",
            "python310",
            "--region",
            "us-central1",
            "--source",
            source_path,
            "--entry-point",
            "defog_query_http",
            "--max-instances",
            "1",
            "--set-env-vars",
            f"DEFOG_CREDS_64={creds64_str}",
            "--trigger-http",
            "--gen2",
            "--allow-unauthenticated",
        ]
        try:
            cmd_str = " ".join(cmd)
            print(f"executing gcloud command:\n{cmd_str}")
            subprocess.check_call(cmd)
            print("gcloud command executed successfully")
        except subprocess.CalledProcessError as e:
            print(f"Error deploying Cloud Function:\n{e}")
    elif cloud_provider == "aws":
        # base64 encode defog credentials for ease of passing around in cli
        creds64_str = df.to_base64_creds()
        # get base config from defog package, add env vars
        base_config_path = os.path.join(defog.__path__[0], "aws", "base_config.json")
        with open(base_config_path, "r") as f:
            chalice_config = json.load(f)
        chalice_config["environment_variables"] = {"DEFOG_CREDS_64": creds64_str}
        chalice_config = parse_update(
            sys.argv[3:], ["app_name", "version"], chalice_config
        )
        aws_path = os.path.join(home_dir, ".defog", "aws")
        chalice_path = os.path.join(aws_path, ".chalice")
        if not os.path.exists(chalice_path):
            print(f"creating {chalice_path}")
            os.makedirs(chalice_path)
        chalice_config_path = os.path.join(chalice_path, "config.json")
        # save to .defog/aws/.chalice/config.json
        with open(chalice_config_path, "w") as f:
            json.dump(chalice_config, f)
        # copy over app.py and requirements.txt to .defog/aws
        app_path = os.path.join(defog.__path__[0], "aws", "app.py")
        req_path = os.path.join(defog.__path__[0], "aws", "requirements.txt")
        shutil.copy(app_path, aws_path)
        shutil.copy(req_path, aws_path)

        # deploy with chalice
        try:
            print("deploying with Chalice...")
            os.chdir(aws_path)
            subprocess.check_call(["chalice", "deploy"])
            os.chdir("../..")
            print("deployed aws lambda successfully with Chalice.")
            print(f"You can find the chalice deployment artifacts in {aws_path}")
        except subprocess.CalledProcessError as e:
            print(f"Error deploying with Chalice:\n{e}")
    else:
        raise ValueError("Cloud provider must be 'gcp' or 'aws'.")


def quota():
    """
    Check your current usage and quota.
    """
    df = defog.Defog()
    resp = df.get_quota()
    if resp:
        if resp["premium"]:
            print(f"You are currently on the premium plan with unrestricted usage.")
        else:
            print(f"You are currently on the free plan with 1000 queries per month.")
        print("Usage so far:")
        print(f"{'Month':<10}{'API Calls':<10}")
        for month, api_calls in resp["data"]:
            print(f"{month}:  {api_calls}")
    else:
        print(f"Failed to get quota")


# helper function to format different field types into strings
def to_str(field) -> str:
    if isinstance(field, str):
        return field
    elif isinstance(field, int):
        return str(field)
    elif isinstance(field, float):
        return str(field)
    elif isinstance(field, datetime.datetime):
        return field.strftime("%Y-%m-%d")
    elif isinstance(field, datetime.date):
        return field.strftime("%Y-%m-%d")
    elif isinstance(field, datetime.timedelta):
        return str(field)
    elif isinstance(field, datetime.time):
        return field.strftime("%H:%M:%S")
    elif isinstance(field, list):
        return str(field)
    elif isinstance(field, dict):
        return str(field)
    elif isinstance(field, bool):
        return str(field)
    elif isinstance(field, decimal.Decimal):
        return str(field)
    elif field is None:
        return "NULL"
    else:
        raise ValueError(f"Unknown type: {type(field)}")


# helper function to print results in tabular format
def print_table(columns, data):
    # Calculate the maximum width of each column, including headers
    data_header = data + [tuple(columns)]
    column_widths = [
        max(len(to_str(row[i])) for row in data_header) for i in range(len(columns))
    ]

    # Print the table headers
    for i, column in enumerate(columns):
        print(column.ljust(column_widths[i]), end=" | ")
    print()

    # Print the table divider
    for i, column_width in enumerate(column_widths):
        print("-" * column_width, end="-+-" if i < len(column_widths) - 1 else "-|\n")

    # Print the table data
    for row in data:
        for i, value in enumerate(row):
            print(to_str(value).ljust(column_widths[i]), end=" | ")
        print()


def serve_webserver():
    import uvicorn

    from defog.serve import app

    uvicorn.run(app, host="localhost", port=1235)


def serve_static():
    import http.server
    import socketserver
    import webbrowser

    port = 8002
    directory = os.path.join(defog.__path__[0], "static")

    class Handler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=directory, **kwargs)

    webbrowser.open(f"http://localhost:{port}")
    with socketserver.TCPServer(("", port), Handler) as httpd:
        print(f"Serving at port {port}")
        print(f"Static folder is {directory}")
        httpd.serve_forever()


def serve():
    """
    Serve a defog server locally.
    """
    print("Starting defog server...")
    print("Serving webserver...")
    webserver_process = subprocess.Popen(["defog", "serve-webserver"])
    time.sleep(2)

    print("Serving static files...")
    static_process = subprocess.Popen(["defog", "serve-static"])

    print("Press Ctrl+C to exit.")
    try:
        while True:
            pass
    except KeyboardInterrupt:
        print("Exiting...")
        static_process.terminate()
        webserver_process.terminate()
        sys.exit(0)


if __name__ == "__main__":
    main()



import os
import time
import json
import traceback
from dataclasses import dataclass
from typing import Dict, List, Optional, Any, Union, Callable

from defog.llm.utils_function_calling import (
    get_function_specs,
    convert_tool_choice,
    execute_tool,
    execute_tool_async,
    verify_post_tool_function,
)
import inspect
import asyncio

LLM_COSTS_PER_TOKEN = {
    "chatgpt-4o": {"input_cost_per1k": 0.0025, "output_cost_per1k": 0.01},
    "gpt-4o": {"input_cost_per1k": 0.0025, "output_cost_per1k": 0.01},
    "gpt-4o-mini": {"input_cost_per1k": 0.00015, "output_cost_per1k": 0.0006},
    "o1": {"input_cost_per1k": 0.015, "output_cost_per1k": 0.06},
    "o1-preview": {"input_cost_per1k": 0.015, "output_cost_per1k": 0.06},
    "o1-mini": {"input_cost_per1k": 0.003, "output_cost_per1k": 0.012},
    "o3-mini": {"input_cost_per1k": 0.0011, "output_cost_per1k": 0.0044},
    "gpt-4-turbo": {"input_cost_per1k": 0.01, "output_cost_per1k": 0.03},
    "gpt-3.5-turbo": {"input_cost_per1k": 0.0005, "output_cost_per1k": 0.0015},
    "claude-3-5-sonnet": {"input_cost_per1k": 0.003, "output_cost_per1k": 0.015},
    "claude-3-5-haiku": {"input_cost_per1k": 0.00025, "output_cost_per1k": 0.00125},
    "claude-3-opus": {"input_cost_per1k": 0.015, "output_cost_per1k": 0.075},
    "claude-3-sonnet": {"input_cost_per1k": 0.003, "output_cost_per1k": 0.015},
    "claude-3-haiku": {"input_cost_per1k": 0.00025, "output_cost_per1k": 0.00125},
    "gemini-1.5-pro": {"input_cost_per1k": 0.00125, "output_cost_per1k": 0.005},
    "gemini-1.5-flash": {"input_cost_per1k": 0.000075, "output_cost_per1k": 0.0003},
    "gemini-1.5-flash-8b": {
        "input_cost_per1k": 0.0000375,
        "output_cost_per1k": 0.00015,
    },
    "gemini-2.0-flash": {
        "input_cost_per1k": 0.00010,
        "output_cost_per1k": 0.0004,
    },
    "deepseek-chat": {
        "input_cost_per1k": 0.00014,
        "output_cost_per1k": 0.00028,
    },
    "deepseek-reasoner": {
        "input_cost_per1k": 0.00055,
        "output_cost_per1k": 0.00219,
    },
}


@dataclass
class LLMResponse:
    content: Any
    model: str
    time: float
    input_tokens: int
    output_tokens: int
    output_tokens_details: Optional[Dict[str, int]] = None
    cost_in_cents: Optional[float] = None
    tools_used: Optional[List[str]] = None
    tool_outputs: Optional[Dict[str, str]] = None

    def __post_init__(self):
        if self.model in LLM_COSTS_PER_TOKEN:
            model_name = self.model
        else:
            # Attempt partial matches if no exact match
            model_name = None
            potential_model_names = []
            for mname in LLM_COSTS_PER_TOKEN.keys():
                if mname in self.model:
                    potential_model_names.append(mname)
            if len(potential_model_names) > 0:
                model_name = max(potential_model_names, key=len)

        if model_name:
            self.cost_in_cents = (
                self.input_tokens
                / 1000
                * LLM_COSTS_PER_TOKEN[model_name]["input_cost_per1k"]
                + self.output_tokens
                / 1000
                * LLM_COSTS_PER_TOKEN[model_name]["output_cost_per1k"]
            ) * 100


#
# --------------------------------------------------------------------------------
# 1) ANTHROPIC
# --------------------------------------------------------------------------------
#


def _build_anthropic_params(
    messages: List[Dict[str, str]],
    model: str,
    max_completion_tokens: int,
    temperature: float,
    stop: List[str],
    tools: List[Callable] = None,
    tool_choice: str = None,
    timeout: int = 100,
):
    """Create the parameter dict for Anthropic's .messages.create()."""
    if len(messages) >= 1 and messages[0].get("role") == "system":
        sys_msg = messages[0]["content"]
        messages = messages[1:]
    else:
        sys_msg = ""

    params = {
        "system": sys_msg,
        "messages": messages,
        "model": model,
        "max_tokens": max_completion_tokens,
        "temperature": temperature,
        "stop_sequences": stop,
        "timeout": timeout,
    }
    if tools:
        function_specs = get_function_specs(tools, model)
        params["tools"] = function_specs
        if tool_choice:
            tool_names_list = [func.__name__ for func in tools]
            tool_choice = convert_tool_choice(tool_choice, tool_names_list, model)
            params["tool_choice"] = tool_choice
        else:
            params["tool_choice"] = {"type": "auto"}

    return params, messages  # returning updated messages in case we want them


async def _process_anthropic_response(
    client,
    response,
    request_params: Dict[str, Any],
    tools: List[Callable],
    tool_dict: Dict[str, Callable],
    is_async: bool,
    post_tool_function: Callable = None,
):
    """
    Extract content (including any tool calls) and usage info from Anthropic response.
    Handles chaining of tool calls.
    """
    from anthropic.types import ToolUseBlock, TextBlock

    if response.stop_reason == "max_tokens":
        raise Exception("Max tokens reached")
    if len(response.content) == 0:
        raise Exception("Max tokens reached")

    # If we have tools, handle dynamic chaining:
    tools_used = []
    tool_outputs = []
    total_input_tokens = 0
    total_output_tokens = 0
    if tools and len(tools) > 0:
        while True:
            total_input_tokens += response.usage.input_tokens
            total_output_tokens += response.usage.output_tokens
            # Check if the response contains a tool call
            tool_call_block = next(
                (
                    block
                    for block in response.content
                    if isinstance(block, ToolUseBlock)
                ),
                None,
            )
            text_block = next(
                (block for block in response.content if isinstance(block, TextBlock)),
                None,
            )
            if tool_call_block:
                try:
                    func_name = tool_call_block.name
                    args = tool_call_block.input
                    tool_id = tool_call_block.id
                except Exception as e:
                    raise Exception(f"Error parsing tool call: {e}")

                try:
                    tool_to_call = tool_dict[func_name]
                except KeyError:
                    raise Exception(f"Tool `{func_name}` not found.")

                # Execute tool depending on whether it is async
                try:
                    if inspect.iscoroutinefunction(tool_to_call):
                        result = await execute_tool_async(tool_to_call, args)
                    else:
                        result = execute_tool(tool_to_call, args)
                except Exception as e:
                    raise Exception(f"Error executing tool `{func_name}`: {e}")

                # Execute post-tool function if provided
                if post_tool_function:
                    if inspect.iscoroutinefunction(post_tool_function):
                        await post_tool_function(
                            function_name=func_name, input_args=args, tool_result=result
                        )
                    else:
                        post_tool_function(
                            function_name=func_name, input_args=args, tool_result=result
                        )

                # Store the tool call, result, and text
                tools_used.append(func_name)
                tool_outputs.append(
                    {
                        "name": func_name,
                        "args": args,
                        "result": result,
                        "text": text_block.text if text_block else None,
                    }
                )

                # Append the tool call as an assistant response
                request_params["messages"].append(
                    {
                        "role": "assistant",
                        "content": [
                            {
                                "type": "tool_use",
                                "id": tool_id,
                                "name": func_name,
                                "input": args,
                            }
                        ],
                    }
                )

                # Append the tool result as a user response
                request_params["messages"].append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "tool_result",
                                "tool_use_id": tool_id,
                                "content": str(result),
                            }
                        ],
                    }
                )

                # Set tool_choice to "auto" so that the next message will be generated normally
                request_params["tool_choice"] = (
                    {"type": "auto"}
                    if request_params["tool_choice"] != "auto"
                    else None
                )

                # Make next call
                if is_async:
                    response = await client.messages.create(**request_params)
                else:
                    response = client.messages.create(**request_params)
            else:
                content = response.content[0].text
                break
    else:
        # No tools provided
        content = response.content[0].text

    usage = response.usage
    total_input_tokens += usage.input_tokens
    total_output_tokens += usage.output_tokens
    return content, tools_used, tool_outputs, total_input_tokens, total_output_tokens


def _process_anthropic_response_handler(
    client,
    response,
    request_params: Dict[str, Any],
    tools: List[Callable],
    tool_dict: Dict[str, Callable],
    is_async: bool = False,
    post_tool_function: Callable = None,
):
    """
    Processes Anthropic's response by determining whether to execute the response handling
    synchronously or asynchronously. This function acts as a wrapper around _process_anthropic_response,
    deciding the execution mode based on the is_async parameter.

    Parameters:
    - client: The client instance used for communication.
    - response: The response object from Anthropic.
    - request_params: A dictionary of request parameters that resulted in the response.
    - tools: A list of callable tools available for function calling.
    - tool_dict: A dictionary mapping tool names to their callable functions.
    - is_async: A boolean flag indicating whether to execute asynchronously.

    Returns:
    - The processed response content, input tokens, and output tokens from the response.
    """
    try:
        if is_async:
            return _process_anthropic_response(
                client=client,
                response=response,
                request_params=request_params,
                tools=tools,
                tool_dict=tool_dict,
                is_async=is_async,
                post_tool_function=post_tool_function,
            )  # Caller must await this
        else:
            return asyncio.run(
                _process_anthropic_response(
                    client=client,
                    response=response,
                    request_params=request_params,
                    tools=tools,
                    tool_dict=tool_dict,
                    is_async=is_async,
                    post_tool_function=post_tool_function,
                )
            )
    except Exception as e:
        raise Exception("Error processing Anthropic response:", e)


def chat_anthropic(
    messages: List[Dict[str, str]],
    model: str = "claude-3-5-sonnet-20241022",
    max_completion_tokens: int = 8192,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    post_tool_function: Callable = None,
):
    """
    Synchronous Anthropic chat.

    Parameters:
    - messages: A list of dictionaries representing the conversation so far.
    - model: The anthropic model to use for the chat.
    - max_completion_tokens: The maximum number of tokens for the completion.
    - temperature: Ranges from 0.0 to 1.0. Defaults to 1.0. Use temperature closer to 0.0 for analytical / multiple choice, and closer to 1.0 for creative and generative tasks.
    - stop: Custom text sequences that will cause the model to stop generating.
    - response_format: NA
    - seed: NA
    - tools: The list of tools the model may call.
    - tool_choice: Controls which (if any) tool is called by the model.
        "auto": calls 0, 1, or multiple functions,
        "required": calls at least one function,
        "<function_name>": calls only the specified function
    - post_tool_function: An optional function to handle tool responses post-generation. The function takes exactly 3 arguments: function_name, input_args, tool_result

    Returns:
    - LLMResponse which contains the response content, input tokens, output tokens, tools used, and tool outputs
    """
    from anthropic import Anthropic

    if post_tool_function:
        verify_post_tool_function(post_tool_function)

    t = time.time()
    client = Anthropic()
    request_params, _ = _build_anthropic_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        tools=tools,
        tool_choice=tool_choice,
    )

    # Construct a tool dict if needed
    tool_dict = {}
    if tools and len(tools) > 0 and "tools" in request_params:
        tool_dict = {tool.__name__: tool for tool in tools}

    response = client.messages.create(**request_params)
    content, tools_used, tool_outputs, input_toks, output_toks = (
        _process_anthropic_response_handler(
            client=client,
            response=response,
            request_params=request_params,
            tools=tools,
            tool_dict=tool_dict,
            is_async=False,
            post_tool_function=post_tool_function,
        )
    )

    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=input_toks,
        output_tokens=output_toks,
        tools_used=tools_used,
        tool_outputs=tool_outputs,
    )


async def chat_anthropic_async(
    messages: List[Dict[str, str]],
    model: str = "claude-3-5-sonnet-20241022",
    max_completion_tokens: int = 8192,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    store=True,
    metadata=None,
    timeout=100,
    prediction=None,
    reasoning_effort=None,
    post_tool_function: Callable = None,
):
    """
    Asynchronous Anthropic chat.

    Parameters:
    - messages: A list of dictionaries representing the conversation so far.
    - model: The anthropic model to use for the chat.
    - max_completion_tokens: The maximum number of tokens for the completion.
    - temperature: Ranges from 0.0 to 1.0. Use temperature closer to 0.0 for analytical / multiple choice, and closer to 1.0 for creative and generative tasks.
    - stop: Custom text sequences that will cause the model to stop generating.
    - response_format: NA
    - seed: NA
    - tools: The list of tools the model may call.
    - tool_choice: Controls which (if any) tool is called by the model.
        "auto": calls 0, 1, or multiple functions,
        "required": calls at least one function,
        "<function_name>": calls only the specified function
    - store: NA
    - metadata: NA
    - timeout: NA
    - prediction: NA
    - reasoning_effort: NA
    - post_tool_function: An optional function to handle tool responses post-generation. The function takes exactly 3 arguments: function_name, input_args, tool_result

    Returns:
    - LLMResponse which contains the response content, input tokens, output tokens, tools used, and tool outputs
    """
    from anthropic import AsyncAnthropic

    t = time.time()
    client = AsyncAnthropic()
    params, _ = _build_anthropic_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        tools=tools,
        tool_choice=tool_choice,
    )

    # Construct a tool dict if needed
    tool_dict = {}
    if tools and len(tools) > 0 and "tools" in params:
        tool_dict = {tool.__name__: tool for tool in tools}

    response = await client.messages.create(**params)
    content, tools_used, tool_outputs, input_toks, output_toks = (
        await _process_anthropic_response_handler(
            client=client,
            response=response,
            request_params=params,
            tools=tools,
            tool_dict=tool_dict,
            is_async=True,
            post_tool_function=post_tool_function,
        )
    )

    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=input_toks,
        output_tokens=output_toks,
        tools_used=tools_used,
        tool_outputs=tool_outputs,
    )


#
# --------------------------------------------------------------------------------
# 2) OPENAI
# --------------------------------------------------------------------------------
#


def _build_openai_params(
    messages: List[Dict[str, str]],
    model: str,
    max_completion_tokens: int,
    temperature: float,
    stop: List[str],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    prediction: Dict[str, str] = None,
    reasoning_effort: str = None,
    store: bool = True,
    metadata: Dict[str, str] = None,
    timeout: int = 100,
):
    """
    Build the parameter dictionary for OpenAI's chat.completions.create().
    Also handles special logic for o1-mini, o1-preview, deepseek-chat, etc.
    """
    # Potentially move system message to user message for certain model families:
    if model in [
        "o1-mini",
        "o1-preview",
        "o1",
        "deepseek-chat",
        "deepseek-reasoner",
        "o3-mini",
    ]:
        sys_msg = None
        for i in range(len(messages)):
            if messages[i].get("role") == "system":
                sys_msg = messages.pop(i)["content"]
                break
        if sys_msg:
            for i in range(len(messages)):
                if messages[i].get("role") == "user":
                    messages[i]["content"] = sys_msg + "\n" + messages[i]["content"]
                    break

    request_params = {
        "messages": messages,
        "model": model,
        "max_completion_tokens": max_completion_tokens,
        "temperature": temperature,
        "stop": stop,
        "seed": seed,
        "store": store,
        "metadata": metadata,
        "timeout": timeout,
    }

    # Tools are only supported for certain models
    if (
        tools
        and len(tools) > 0
        and model
        not in [
            "o1-mini",
            "o1-preview",
            "deepseek-chat",
            "deepseek-reasoner",
        ]
    ):
        function_specs = get_function_specs(tools, model)
        request_params["tools"] = function_specs
        if tool_choice:
            tool_names_list = [func.__name__ for func in tools]
            tool_choice = convert_tool_choice(tool_choice, tool_names_list, model)
            request_params["tool_choice"] = tool_choice
        else:
            request_params["tool_choice"] = "auto"

        # only set parallel_tool_calls for gpt-4o based models
        # not supported in o models
        if model in ["gpt-4o", "gpt-4o-mini"]:
            request_params["parallel_tool_calls"] = False

    # Some models do not allow temperature or response_format:
    if model.startswith("o") or model == "deepseek-reasoner":
        request_params.pop("temperature", None)
    if model in ["o1-mini", "o1-preview", "deepseek-chat", "deepseek-reasoner"]:
        request_params.pop("response_format", None)

    # Reasoning effort
    if model.startswith("o") and reasoning_effort is not None:
        request_params["reasoning_effort"] = reasoning_effort

    # Special case: model in ["gpt-4o", "gpt-4o-mini"] with `prediction`
    if model in ["gpt-4o", "gpt-4o-mini"] and prediction is not None:
        request_params["prediction"] = prediction
        request_params.pop("max_completion_tokens", None)
        request_params.pop("response_format", None)

    # Finally, set response_format if still relevant:
    if response_format:
        request_params["response_format"] = response_format
        # cannot have stop when using response_format
        request_params.pop("stop", None)

    return request_params


async def _process_openai_response(
    client,
    response,
    request_params: Dict[str, Any],
    tools: List[Callable],
    tool_dict: Dict[str, Callable],
    response_format,
    model: str,
    is_async: bool,
    post_tool_function: Callable = None,
):
    """
    Extract content (including any tool calls) and usage info from OpenAI response.
    Handles chaining of tool calls.
    """
    if len(response.choices) == 0:
        raise Exception("No response from OpenAI.")
    if response.choices[0].finish_reason == "length":
        raise Exception("Max tokens reached")

    # If we have tools, handle dynamic chaining:
    tools_used = []
    tool_outputs = []
    total_input_tokens = 0
    total_output_tokens = 0
    if tools and len(tools) > 0:
        while True:
            total_input_tokens += response.usage.prompt_tokens
            total_output_tokens += response.usage.completion_tokens
            message = response.choices[0].message
            if message.tool_calls:
                tool_call = message.tool_calls[0]
                func_name = tool_call.function.name
                try:
                    args = json.loads(tool_call.function.arguments)
                except json.JSONDecodeError:
                    args = {}

                try:
                    tool_to_call = tool_dict[func_name]
                except KeyError:
                    raise Exception(f"Tool `{func_name}` not found")

                # Execute tool depending on whether it is async
                try:
                    if inspect.iscoroutinefunction(tool_to_call):
                        result = await execute_tool_async(tool_to_call, args)
                    else:
                        result = execute_tool(tool_to_call, args)
                except Exception as e:
                    raise Exception(f"Error executing tool `{func_name}`: {e}")

                # Execute post-tool function if provided
                if post_tool_function:
                    if inspect.iscoroutinefunction(post_tool_function):
                        await post_tool_function(
                            function_name=func_name, input_args=args, tool_result=result
                        )
                    else:
                        post_tool_function(
                            function_name=func_name, input_args=args, tool_result=result
                        )

                # Store the tool call, result, and text
                tools_used.append(func_name)
                tool_outputs.append(
                    {
                        "name": func_name,
                        "args": args,
                        "result": result,
                        "text": message.content if message.content else None,
                    }
                )

                # Append the tool calls as an assistant response
                request_params["messages"].append(
                    {
                        "role": "assistant",
                        "tool_calls": message.tool_calls,
                    }
                )

                # Append the tool message
                request_params["messages"].append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "content": str(result),
                    }
                )

                # Set tool_choice to "auto" so that the next message will be generated normally
                request_params["tool_choice"] = (
                    "auto" if request_params["tool_choice"] != "auto" else None
                )

                # Make next call
                if is_async:
                    response = await client.chat.completions.create(**request_params)
                else:
                    response = client.chat.completions.create(**request_params)
            else:
                content = message.content
                break
    else:
        # No tools provided
        if response_format and model not in ["o1-mini", "o1-preview"]:
            content = response.choices[0].message.parsed
        else:
            content = response.choices[0].message.content

    usage = response.usage
    total_input_tokens += usage.prompt_tokens
    total_output_tokens += usage.completion_tokens
    return (
        content,
        tools_used,
        tool_outputs,
        total_input_tokens,
        total_output_tokens,
        usage.completion_tokens_details,
    )


def _process_openai_response_handler(
    client,
    response,
    request_params: Dict[str, Any],
    tools: List[Callable],
    tool_dict: Dict[str, Callable],
    response_format,
    model: str,
    is_async: bool = False,
    post_tool_function: Callable = None,
):
    """
    Processes OpenAI's response by determining whether to execute the response handling
    synchronously or asynchronously. This function acts as a wrapper around _process_openai_response,
    deciding the execution mode based on the is_async parameter.

    Parameters:
    - client: The client instance used for communication.
    - response: The response object from OpenAI.
    - request_params: A dictionary of request parameters that resulted in the response.
    - tools: A list of callable tools available for function calling.
    - tool_dict: A dictionary mapping tool names to their callable functions.
    - is_async: A boolean flag indicating whether to execute asynchronously.

    Returns:
    - The processed response content, input tokens, and output tokens from the response.
    """
    try:
        if is_async:
            return _process_openai_response(
                client=client,
                response=response,
                request_params=request_params,
                tools=tools,
                tool_dict=tool_dict,
                response_format=response_format,
                model=model,
                is_async=is_async,
                post_tool_function=post_tool_function,
            )  # Caller must await this
        else:
            return asyncio.run(
                _process_openai_response(
                    client=client,
                    response=response,
                    request_params=request_params,
                    tools=tools,
                    tool_dict=tool_dict,
                    response_format=response_format,
                    model=model,
                    is_async=is_async,
                    post_tool_function=post_tool_function,
                )
            )

    except Exception as e:
        raise Exception("Error processing OpenAI response:", e)


def chat_openai(
    messages: List[Dict[str, str]],
    model: str = "gpt-4o",
    max_completion_tokens: int = 16384,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    base_url: str = "https://api.openai.com/v1/",
    api_key: str = os.environ.get("OPENAI_API_KEY", ""),
    prediction: Dict[str, str] = None,
    reasoning_effort: str = None,
    store: bool = True,
    metadata: Dict[str, str] = None,
    timeout: int = 100,
    post_tool_function: Callable = None,
):
    """
    Synchronous OpenAI chat.

    Parameters:
    - messages: The list of messages to send to the LLM.
    - model: The OpenAI model to use for the chat.
    - max_completion_tokens: The maximum number of tokens to return in the response.
    - temperature: Between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.
    - stop: Up to 4 sequences where the API will stop generating further tokens.
    - response_format: The format that the model must output.
    - seed: If specified, OpenAI will try their best to sample deterministically
    - tools: The list of tools the model may call.
    - tool_choice: Controls which (if any) tool is called by the model.
        "auto": calls 0, 1, or multiple functions,
        "required": calls at least one function,
        "<function_name>": calls only the specified function
    - base_url: The base URL to use for the chat.
    - api_key: The OpenAI API key
    - prediction: Configuration for a Predicted Output.
    - reasoning_effort: "low", "medium", or "high". Only for o1 and o3 models
    - store: Whether or not to store the output of this chat completion request for use in model distillation or evals products.
    - metadata: Set of 16 key-value pairs that can be attached to an object. This can be useful for storing additional information about the object in a structured format, and querying for objects via API or the dashboard.
    - timeout: No. of seconds before the request times out.
    - post_tool_function: An optional function to handle tool responses post-generation. The function takes exactly 3 arguments: function_name, input_args, tool_result

    Returns:
    - LLMResponse which contains the response content, input tokens, output tokens, output token details, tools used, and tool outputs
    """
    from openai import OpenAI

    if post_tool_function:
        verify_post_tool_function(post_tool_function)

    t = time.time()
    client_openai = OpenAI(base_url=base_url, api_key=api_key)
    request_params = _build_openai_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        response_format=response_format,
        seed=seed,
        tools=tools,
        tool_choice=tool_choice,
        store=store,
        metadata=metadata,
        timeout=timeout,
        prediction=prediction,
        reasoning_effort=reasoning_effort,
    )

    # Construct a tool dict if needed
    tool_dict = {}
    if tools and len(tools) > 0 and "tools" in request_params:
        tool_dict = {tool.__name__: tool for tool in tools}

    # If response_format is set, we do parse
    if request_params.get("response_format"):
        response = client_openai.beta.chat.completions.parse(**request_params)
    else:
        response = client_openai.chat.completions.create(**request_params)

    (
        content,
        tools_used,
        tool_outputs,
        prompt_tokens,
        output_tokens,
        completion_token_details,
    ) = _process_openai_response_handler(
        client=client_openai,
        response=response,
        request_params=request_params,
        tools=tools,
        tool_dict=tool_dict,
        response_format=response_format,
        model=model,
        is_async=False,
        post_tool_function=post_tool_function,
    )

    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=prompt_tokens,
        output_tokens=output_tokens,
        output_tokens_details=completion_token_details,
        tools_used=tools_used,
        tool_outputs=tool_outputs,
    )


async def chat_openai_async(
    messages: List[Dict[str, str]],
    model: str = "gpt-4o",
    max_completion_tokens: int = 16384,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    store: bool = True,
    metadata: Dict[str, str] = None,
    timeout: int = 100,
    base_url: str = "https://api.openai.com/v1/",
    api_key: str = os.environ.get("OPENAI_API_KEY", ""),
    prediction: Dict[str, str] = None,
    reasoning_effort: str = None,
    post_tool_function: Callable = None,
):
    """
    Asynchronous OpenAI chat.

    Parameters:
    - messages: The list of messages to send to the LLM.
    - model: The OpenAI model to use for the chat.
    - max_completion_tokens: The maximum number of tokens to return in the response.
    - temperature: Between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.
    - stop: Up to 4 sequences where the API will stop generating further tokens.
    - response_format: The format that the model must output.
    - seed: If specified, OpenAI will try their best to sample deterministically
    - tools: The list of tools the model may call.
    - tool_choice: Controls which (if any) tool is called by the model.
        "auto": calls 0, 1, or multiple functions,
        "required": calls at least one function,
        "<function_name>": calls only the specified function
    - base_url: The base URL to use for the chat.
    - api_key: The OpenAI API key
    - prediction: Configuration for a Predicted Output.
    - reasoning_effort: "low", "medium", or "high". Only for o1 and o3 models
    - store: Whether or not to store the output of this chat completion request for use in model distillation or evals products.
    - metadata: Set of 16 key-value pairs that can be attached to an object. This can be useful for storing additional information about the object in a structured format, and querying for objects via API or the dashboard.
    - timeout: No. of seconds before the request times out.
    - post_tool_function: An optional function to handle tool responses post-generation. The function takes exactly 3 arguments: function_name, input_args, tool_result

    Returns:
    - LLMResponse which contains the response content, input tokens, output tokens, output token details, tools used, and tool outputs
    """
    from openai import AsyncOpenAI

    t = time.time()
    client_openai = AsyncOpenAI(base_url=base_url, api_key=api_key)
    request_params = _build_openai_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        response_format=response_format,
        seed=seed,
        tools=tools,
        tool_choice=tool_choice,
        prediction=prediction,
        reasoning_effort=reasoning_effort,
        store=store,
        metadata=metadata,
        timeout=timeout,
    )

    # Build a tool dict if needed
    tool_dict = {}
    if tools and len(tools) > 0 and "tools" in request_params:
        tool_dict = {tool.__name__: tool for tool in tools}

    # If response_format is set, we do parse
    if request_params.get("response_format"):
        response = await client_openai.beta.chat.completions.parse(**request_params)
    else:
        response = await client_openai.chat.completions.create(**request_params)

    (
        content,
        tools_used,
        tool_outputs,
        prompt_tokens,
        output_tokens,
        completion_token_details,
    ) = await _process_openai_response_handler(
        client=client_openai,
        response=response,
        request_params=request_params,
        tools=tools,
        tool_dict=tool_dict,
        response_format=response_format,
        model=model,
        is_async=True,
        post_tool_function=post_tool_function,
    )

    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=prompt_tokens,
        output_tokens=output_tokens,
        output_tokens_details=completion_token_details,
        tools_used=tools_used,
        tool_outputs=tool_outputs,
    )


#
# --------------------------------------------------------------------------------
# 3) TOGETHER
# --------------------------------------------------------------------------------
#


def _build_together_params(
    messages: List[Dict[str, str]],
    model: str,
    max_completion_tokens: int,
    temperature: float,
    stop: List[str],
    seed: int = 0,
):
    return {
        "messages": messages,
        "model": model,
        "max_tokens": max_completion_tokens,
        "temperature": temperature,
        "stop": stop,
        "seed": seed,
    }


def _process_together_response(response):
    if response.choices[0].finish_reason == "length":
        raise Exception("Max tokens reached")
    if len(response.choices) == 0:
        raise Exception("Max tokens reached")
    return (
        response.choices[0].message.content,
        response.usage.prompt_tokens,
        response.usage.completion_tokens,
    )


def chat_together(
    messages: List[Dict[str, str]],
    model: str = "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
    max_completion_tokens: int = 4096,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools=None,
    tool_choice=None,
    post_tool_function: Callable = None,
):
    """Synchronous Together chat."""
    from together import Together

    t = time.time()
    client_together = Together()
    params = _build_together_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        seed=seed,
    )
    response = client_together.chat.completions.create(**params)

    content, input_toks, output_toks = _process_together_response(response)
    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=input_toks,
        output_tokens=output_toks,
    )


async def chat_together_async(
    messages: List[Dict[str, str]],
    model: str = "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
    max_completion_tokens: int = 4096,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools=None,
    tool_choice=None,
    store=True,
    metadata=None,
    timeout=100,
    prediction=None,
    reasoning_effort=None,
    post_tool_function: Callable = None,
):
    """Asynchronous Together chat."""
    from together import AsyncTogether

    t = time.time()
    client_together = AsyncTogether(timeout=timeout)
    params = _build_together_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        seed=seed,
    )
    response = await client_together.chat.completions.create(**params)

    content, input_toks, output_toks = _process_together_response(response)
    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=input_toks,
        output_tokens=output_toks,
    )


#
# --------------------------------------------------------------------------------
# 4) GEMINI
# --------------------------------------------------------------------------------
#


def _build_gemini_params(
    messages: List[Dict[str, str]],
    model: str,
    max_completion_tokens: int,
    temperature: float,
    stop: List[str],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
):
    """Construct parameters for Gemini's generate_content call."""

    from google.genai import types

    if messages[0]["role"] == "system":
        system_msg = messages[0]["content"]
        messages = messages[1:]
    else:
        system_msg = None

    # Combine all user/assistant messages into one string and create a types.Content object
    messages_str = "\n".join([m["content"] for m in messages])
    user_prompt_content = types.Content(
        role="user",
        parts=[types.Part.from_text(text=messages_str)],
    )
    messages = [user_prompt_content]
    request_params = {
        "temperature": temperature,
        "system_instruction": system_msg,
        "max_output_tokens": max_completion_tokens,
        "stop_sequences": stop,
    }

    if tools:
        function_specs = get_function_specs(tools, model)
        request_params["tools"] = function_specs

        # Set up automatic_function_calling and tool_config based on tool_choice
        if tool_choice:
            tool_names_list = [func.__name__ for func in tools]
            tool_choice = convert_tool_choice(tool_choice, tool_names_list, model)
        if tool_choice:
            request_params["automatic_function_calling"] = (
                types.AutomaticFunctionCallingConfig(disable=True)
            )
            request_params["tool_config"] = tool_choice

    if response_format:
        # If we want a JSON / Pydantic format
        # "response_schema" is only recognized if the google.genai library supports it
        request_params["response_mime_type"] = "application/json"
        request_params["response_schema"] = response_format

    return messages, request_params


async def _process_gemini_response(
    client,
    response,
    request_params: Dict[str, Any],
    messages: List[Dict[str, str]],
    tools: List[Callable],
    tool_dict: Dict[str, Callable],
    response_format,
    model: str,
    is_async: bool,
    post_tool_function: Callable = None,
):
    """Extract content (including any tool calls) and usage info from Gemini response.
    Handles chaining of tool calls.
    """

    from google.genai import types

    if len(response.candidates) == 0:
        raise Exception("No response from Gemini.")
    if response.candidates[0].finish_reason == "MAX_TOKENS":
        raise Exception("Max tokens reached")

    # If we have tools, handle dynamic chaining:
    tools_used = []
    tool_outputs = []
    total_input_tokens = 0
    total_output_tokens = 0
    if tools and len(tools) > 0:
        while True:
            total_input_tokens += response.usage_metadata.prompt_token_count
            total_output_tokens += response.usage_metadata.candidates_token_count
            if response.function_calls:
                tool_call = response.function_calls[0]
                func_name = tool_call.name
                args = tool_call.args

                try:
                    tool_to_call = tool_dict[func_name]
                except KeyError:
                    raise Exception(f"Tool `{func_name}` not found")

                # Execute tool depending on whether it is async
                try:
                    if inspect.iscoroutinefunction(tool_to_call):
                        result = await execute_tool_async(tool_to_call, args)
                    else:
                        result = execute_tool(tool_to_call, args)
                except Exception as e:
                    raise Exception(f"Error executing tool `{func_name}`: {e}")

                # Execute post-tool function if provided
                if post_tool_function:
                    if inspect.iscoroutinefunction(post_tool_function):
                        await post_tool_function(
                            function_name=func_name, input_args=args, tool_result=result
                        )
                    else:
                        post_tool_function(
                            function_name=func_name, input_args=args, tool_result=result
                        )

                # Store the tool call, result, and text
                tools_used.append(func_name)
                try:
                    text = response.text
                except Exception as e:
                    text = None
                tool_outputs.append(
                    {
                        "name": func_name,
                        "args": args,
                        "result": result,
                        "text": text,
                    }
                )

                # Append the tool call content to messages
                tool_call_content = response.candidates[0].content
                messages.append(tool_call_content)

                # Append the tool result to messages
                messages.append(
                    types.Content(
                        role="tool",
                        parts=[
                            types.Part.from_function_response(
                                name=func_name,
                                response={"result": str(result)},
                            )
                        ],
                    )
                )

                # Set tool_choice to AUTO so that the next message will be generated normally without required tool calls
                request_params["automatic_function_calling"] = (
                    types.AutomaticFunctionCallingConfig(disable=False)
                )
                request_params["tool_config"] = types.ToolConfig(
                    function_calling_config=types.FunctionCallingConfig(mode="AUTO")
                )

                # Make next call
                if is_async:
                    response = await client.aio.models.generate_content(
                        model=model,
                        contents=messages,
                        config=types.GenerateContentConfig(**request_params),
                    )
                else:
                    response = client.models.generate_content(
                        model=model,
                        contents=messages,
                        config=types.GenerateContentConfig(**request_params),
                    )
            else:
                content = response.text.strip() if response.text else None
                break
    else:
        # No tools provided
        if response_format:
            # Attempt to parse with pydantic model
            content = response_format.model_validate_json(response.text)
        else:
            content = response.text.strip() if response.text else None

    usage = response.usage_metadata
    total_input_tokens += usage.prompt_token_count
    total_output_tokens += usage.candidates_token_count
    return (
        content,
        tools_used,
        tool_outputs,
        total_input_tokens,
        total_output_tokens,
    )


def _process_gemini_response_handler(
    client,
    response,
    request_params: Dict[str, Any],
    messages: List,
    tools: List[Callable],
    tool_dict: Dict[str, Callable],
    response_format,
    model: str,
    is_async=False,
    post_tool_function: Callable = None,
):
    """
    Processes Gemini's response by determining whether to execute the response handling
    synchronously or asynchronously. This function acts as a wrapper around _process_gemini_response,
    deciding the execution mode based on the is_async parameter.

    Parameters:
    - client: The client instance used for communication.
    - response: The response object from Gemini.
    - request_params: A dictionary of request parameters that resulted in the response.
    - tools: A list of callable tools available for function calling.
    - tool_dict: A dictionary mapping tool names to their callable functions.
    - is_async: A boolean flag indicating whether to execute asynchronously.

    Returns:
    - The processed response content, input tokens, and output tokens from the response.
    """
    try:
        if is_async:
            return _process_gemini_response(
                client=client,
                response=response,
                request_params=request_params,
                messages=messages,
                tools=tools,
                tool_dict=tool_dict,
                response_format=response_format,
                model=model,
                is_async=is_async,
                post_tool_function=post_tool_function,
            )  # Caller must await this
        else:
            return asyncio.run(
                _process_gemini_response(
                    client=client,
                    response=response,
                    request_params=request_params,
                    messages=messages,
                    tools=tools,
                    tool_dict=tool_dict,
                    response_format=response_format,
                    model=model,
                    is_async=is_async,
                    post_tool_function=post_tool_function,
                )
            )

    except Exception as e:
        raise Exception("Error processing Gemini response:", e)


def chat_gemini(
    messages: List[Dict[str, str]],
    model: str = "gemini-2.0-flash",
    max_completion_tokens: int = 8192,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    post_tool_function: Callable = None,
):
    """
    Synchronous Gemini chat.

    Parameters:
    - messages: The list of messages to send to the LLM.
    - model: The Gemini model to use for the chat.
    - max_completion_tokens: The maximum number of tokens to return in the response.
    - temperature: Higher values will make the output more random, while lower values like will make it more focused and deterministic.
        Between 0 to 2: gemini-1.5-flash, gemini-1.5-pro, gemini-1.0-pro-002, gemini-2.0-flash
        Between 0 to 1: gemini-1.0-pro-vision, gemini-1.0-pro-001
    - stop: List of strings that will stop the model from generating further tokens.
    - response_format: The format that the model must output. See https://ai.google.dev/gemini-api/docs/structured-output?lang=python
    - seed: NA
    - tools: The list of tools the model may call.
    - tool_choice: Controls which (if any) tool is called by the model.
        "auto": calls 0, 1, or multiple functions,
        "required": calls at least one function,
        "<function_name>": calls only the specified function
    - post_tool_function: An optional function to handle tool responses post-generation. The function takes exactly 3 arguments: function_name, input_args, tool_result

    Returns:
    - LLMResponse which contains the response content, input tokens, output tokens, tools used, and tool outputs
    """
    from google import genai
    from google.genai import types

    if post_tool_function:
        verify_post_tool_function(post_tool_function)

    t = time.time()
    client = genai.Client(api_key=os.getenv("GEMINI_API_KEY", ""))
    messages, request_params = _build_gemini_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        response_format=response_format,
        seed=seed,
        tools=tools,
        tool_choice=tool_choice,
    )

    # Construct a tool dict if needed
    tool_dict = {}
    if tools and len(tools) > 0 and "tools" in request_params:
        tool_dict = {tool.__name__: tool for tool in tools}

    try:
        response = client.models.generate_content(
            model=model,
            contents=messages,
            config=types.GenerateContentConfig(**request_params),
        )
    except Exception as e:
        raise Exception(f"An error occurred: {e}")

    content, tools_used, tool_outputs, input_toks, output_toks = (
        _process_gemini_response_handler(
            client=client,
            response=response,
            request_params=request_params,
            tools=tools,
            tool_dict=tool_dict,
            response_format=response_format,
            messages=messages,
            model=model,
            is_async=False,
            post_tool_function=post_tool_function,
        )
    )
    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=input_toks,
        output_tokens=output_toks,
        tools_used=tools_used,
        tool_outputs=tool_outputs,
    )


async def chat_gemini_async(
    messages: List[Dict[str, str]],
    model: str = "gemini-2.0-flash",
    max_completion_tokens: int = 8192,
    temperature: float = 0.0,
    stop: List[str] = [],
    response_format=None,
    seed: int = 0,
    tools: List[Callable] = None,
    tool_choice: str = None,
    store=True,
    metadata=None,
    timeout=100,
    prediction=None,
    reasoning_effort=None,
    post_tool_function: Callable = None,
):
    """
    Asynchronous Gemini chat.

    Parameters:
    - messages: The list of messages to send to the LLM.
    - model: The Gemini model to use for the chat.
    - max_completion_tokens: The maximum number of tokens to return in the response.
    - temperature: Higher values will make the output more random, while lower values like will make it more focused and deterministic.
        Between 0 to 2: gemini-1.5-flash, gemini-1.5-pro, gemini-1.0-pro-002, gemini-2.0-flash
        Between 0 to 1: gemini-1.0-pro-vision, gemini-1.0-pro-001
    - stop: List of strings that will stop the model from generating further tokens.
    - response_format: The format that the model must output. See https://ai.google.dev/gemini-api/docs/structured-output?lang=python
    - seed: NA
    - tools: The list of tools the model may call.
    - tool_choice: Controls which (if any) tool is called by the model.
        "auto": calls 0, 1, or multiple functions,
        "required": calls at least one function,
        "<function_name>": calls only the specified function
    - store: NA
    - metadata: NA
    - timeout: NA
    - prediction: NA
    - reasoning_effort: NA
    - post_tool_function: An optional function to handle tool responses post-generation. The function takes exactly 3 arguments: function_name, input_args, tool_result

    Returns:
    - LLMResponse which contains the response content, input tokens, output tokens, tools used, and tool outputs
    """
    from google import genai
    from google.genai import types

    t = time.time()
    client = genai.Client(api_key=os.getenv("GEMINI_API_KEY", ""))
    messages, request_params = _build_gemini_params(
        messages=messages,
        model=model,
        max_completion_tokens=max_completion_tokens,
        temperature=temperature,
        stop=stop,
        response_format=response_format,
        seed=seed,
        tools=tools,
        tool_choice=tool_choice,
    )

    # Construct a tool dict if needed
    tool_dict = {}
    if tools and len(tools) > 0 and "tools" in request_params:
        tool_dict = {tool.__name__: tool for tool in tools}

        # Set up automatic_function_calling and tool_config based on tool_choice
        if tool_choice:
            tool_names_list = [func.__name__ for func in tools]
            tool_choice = convert_tool_choice(tool_choice, tool_names_list, model)
        if tool_choice:
            request_params["automatic_function_calling"] = (
                types.AutomaticFunctionCallingConfig(disable=True)
            )
            request_params["tool_config"] = tool_choice

    try:
        response = await client.aio.models.generate_content(
            model=model,
            contents=messages,
            config=types.GenerateContentConfig(**request_params),
        )
    except Exception as e:
        raise Exception(f"An error occurred: {e}")

    content, tools_used, tool_outputs, input_toks, output_toks = (
        await _process_gemini_response_handler(
            client=client,
            response=response,
            request_params=request_params,
            messages=messages,
            tools=tools,
            tool_dict=tool_dict,
            response_format=response_format,
            model=model,
            is_async=True,
            post_tool_function=post_tool_function,
        )
    )
    return LLMResponse(
        model=model,
        content=content,
        time=round(time.time() - t, 3),
        input_tokens=input_toks,
        output_tokens=output_toks,
        tools_used=tools_used,
        tool_outputs=tool_outputs,
    )


def map_model_to_chat_fn_async(model: str) -> Callable:
    """
    Returns the appropriate chat function based on the model.
    """
    if model.startswith("claude"):
        return chat_anthropic_async
    if model.startswith("gemini"):
        return chat_gemini_async
    if (
        model.startswith("gpt")
        or model.startswith("o1")
        or model.startswith("chatgpt")
        or model.startswith("o3")
    ):
        return chat_openai_async
    if model.startswith("deepseek"):
        return chat_openai_async
    if (
        model.startswith("meta-llama")
        or model.startswith("mistralai")
        or model.startswith("Qwen")
    ):
        return chat_together_async
    raise ValueError(f"Unknown model: {model}")


async def chat_async(
    model,
    messages,
    max_completion_tokens=4096,
    temperature=0.0,
    stop=[],
    response_format=None,
    seed=0,
    store=True,
    metadata=None,
    timeout=100,  # in seconds
    backup_model=None,
    prediction=None,
    reasoning_effort=None,
    tools=None,
    tool_choice=None,
    max_retries=3,
    post_tool_function: Callable = None,
) -> LLMResponse:
    """
    Returns the response from the LLM API for a single model that is passed in.
    Includes retry logic with exponential backoff for up to 3 attempts.
    """
    llm_function = map_model_to_chat_fn_async(model)
    base_delay = 1  # Initial delay in seconds

    if post_tool_function:
        verify_post_tool_function(post_tool_function)

    for attempt in range(max_retries):
        try:
            if attempt > 0 and backup_model is not None:
                # For the first attempt, use the original model
                # For subsequent attempts, use the backup model if it is provided
                model = backup_model
                llm_function = map_model_to_chat_fn_async(model)
            if not model.startswith("deepseek"):
                return await llm_function(
                    model=model,
                    messages=messages,
                    max_completion_tokens=max_completion_tokens,
                    temperature=temperature,
                    stop=stop,
                    response_format=response_format,
                    seed=seed,
                    tools=tools,
                    tool_choice=tool_choice,
                    store=store,
                    metadata=metadata,
                    timeout=timeout,
                    prediction=prediction,
                    reasoning_effort=reasoning_effort,
                    post_tool_function=post_tool_function,
                )
            else:
                if not os.getenv("DEEPSEEK_API_KEY"):
                    raise Exception("DEEPSEEK_API_KEY is not set")
                return await llm_function(
                    model=model,
                    messages=messages,
                    max_completion_tokens=max_completion_tokens,
                    temperature=temperature,
                    stop=stop,
                    response_format=response_format,
                    seed=seed,
                    store=store,
                    metadata=metadata,
                    timeout=timeout,
                    prediction=prediction,
                    reasoning_effort=reasoning_effort,
                    base_url="https://api.deepseek.com",
                    api_key=os.getenv("DEEPSEEK_API_KEY"),
                    post_tool_function=post_tool_function,
                )
        except Exception as e:
            delay = base_delay * (2**attempt)  # Exponential backoff
            print(
                f"Attempt {attempt + 1} failed. Retrying in {delay} seconds...",
                flush=True,
            )
            print(f"Error: {e}", flush=True)
            error_trace = traceback.format_exc()
            await asyncio.sleep(delay)

    # If we get here, all attempts failed
    raise Exception(
        "All attempts at calling the chat_async function failed. The latest error traceback was: ",
        error_trace,
    )
